# Copyright 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import random

import netaddr
from neutron_lib import constants as n_const
from oslo_utils import uuidutils
from pyroute2.iproute import linux as iproute_linux
from pyroute2.netlink import rtnl
from pyroute2.netlink.rtnl import ifaddrmsg

from ovn_bgp_agent import constants
from ovn_bgp_agent import exceptions as agent_exc
from ovn_bgp_agent.privileged import linux_net
from ovn_bgp_agent.tests.functional import base as base_functional
from ovn_bgp_agent.tests import utils as test_utils
from ovn_bgp_agent.utils import common as common_utils
from ovn_bgp_agent.utils import linux_net as l_net


IP_ADDRESS_EVENTS = {'RTM_NEWADDR': 'added',
                     'RTM_DELADDR': 'removed'}
IP_ADDRESS_SCOPE = {rtnl.rtscopes['RT_SCOPE_UNIVERSE']: 'global',
                    rtnl.rtscopes['RT_SCOPE_SITE']: 'site',
                    rtnl.rtscopes['RT_SCOPE_LINK']: 'link',
                    rtnl.rtscopes['RT_SCOPE_HOST']: 'host'}


def set_up(ifname):
    linux_net.set_link_attribute(ifname, state='up')


def ip_to_cidr(ip, prefix=None):
    """Convert an ip with no prefix to cidr notation

    :param ip: An ipv4 or ipv6 address.  Convertible to netaddr.IPNetwork.
    :param prefix: Optional prefix.  If None, the default 32 will be used for
        ipv4 and 128 for ipv6.
    """
    net = netaddr.IPNetwork(ip)
    if prefix is not None:
        # Can't pass ip and prefix separately.  Must concatenate strings.
        net = netaddr.IPNetwork(str(net.ip) + '/' + str(prefix))
    return str(net)


def _parse_ip_address(pyroute2_address, device_name):
    ip = linux_net.get_attr(pyroute2_address, 'IFA_ADDRESS')
    ip_length = pyroute2_address['prefixlen']
    event = IP_ADDRESS_EVENTS.get(pyroute2_address.get('event'))
    cidr = ip_to_cidr(ip, prefix=ip_length)
    flags = linux_net.get_attr(pyroute2_address, 'IFA_FLAGS')
    dynamic = not bool(flags & ifaddrmsg.IFA_F_PERMANENT)
    tentative = bool(flags & ifaddrmsg.IFA_F_TENTATIVE)
    dadfailed = bool(flags & ifaddrmsg.IFA_F_DADFAILED)
    scope = IP_ADDRESS_SCOPE[pyroute2_address['scope']]
    return {'name': device_name,
            'cidr': cidr,
            'scope': scope,
            'broadcast': linux_net.get_attr(pyroute2_address, 'IFA_BROADCAST'),
            'dynamic': dynamic,
            'tentative': tentative,
            'dadfailed': dadfailed,
            'event': event}


def get_ip_addresses(ifname):
    device = get_devices_info(ifname=ifname)
    if not device:
        return
    ip_addresses = linux_net.get_ip_addresses(
        index=list(device.values())[0]['index'])
    return [_parse_ip_address(_ip, ifname) for _ip in ip_addresses]


def get_devices_info(**kwargs):
    devices = linux_net.get_link_devices(**kwargs)
    retval = {}
    for device in devices:
        ret = {'index': device['index'],
               'name': linux_net.get_attr(device, 'IFLA_IFNAME'),
               'operstate': linux_net.get_attr(device, 'IFLA_OPERSTATE'),
               'state': device['state'],
               'linkmode': linux_net.get_attr(device, 'IFLA_LINKMODE'),
               'mtu': linux_net.get_attr(device, 'IFLA_MTU'),
               'promiscuity': linux_net.get_attr(device, 'IFLA_PROMISCUITY'),
               'mac': linux_net.get_attr(device, 'IFLA_ADDRESS'),
               'broadcast': linux_net.get_attr(device, 'IFLA_BROADCAST'),
               'master': linux_net.get_attr(device, 'IFLA_MASTER'),
               }
        ifla_link = linux_net.get_attr(device, 'IFLA_LINK')
        if ifla_link:
            ret['parent_index'] = ifla_link
        ifla_linkinfo = linux_net.get_attr(device, 'IFLA_LINKINFO')
        if ifla_linkinfo:
            ret['kind'] = linux_net.get_attr(ifla_linkinfo, 'IFLA_INFO_KIND')
            ret['slave_kind'] = linux_net.get_attr(ifla_linkinfo,
                                                   'IFLA_INFO_SLAVE_KIND')
            ifla_data = linux_net.get_attr(ifla_linkinfo, 'IFLA_INFO_DATA')
            if ret['kind'] == 'vxlan':
                ret['vxlan_id'] = linux_net.get_attr(ifla_data,
                                                     'IFLA_VXLAN_ID')
                ret['vxlan_group'] = linux_net.get_attr(ifla_data,
                                                        'IFLA_VXLAN_GROUP')
                ret['vxlan_link_index'] = linux_net.get_attr(ifla_data,
                                                             'IFLA_VXLAN_LINK')
                ret['vxlan_port'] = linux_net.get_attr(ifla_data,
                                                       'IFLA_VXLAN_PORT')
                ret['vxlan_local'] = linux_net.get_attr(ifla_data,
                                                        'IFLA_VXLAN_LOCAL')
                ret['vxlan_learning'] = bool(
                    linux_net.get_attr(ifla_data, 'IFLA_VXLAN_LEARNING'))
            elif ret['kind'] == 'vlan':
                ret['vlan_id'] = linux_net.get_attr(ifla_data, 'IFLA_VLAN_ID')
            elif ret['kind'] == 'bridge':
                ret['stp'] = linux_net.get_attr(ifla_data, 'IFLA_BR_STP_STATE')
                ret['forward_delay'] = linux_net.get_attr(
                    ifla_data, 'IFLA_BR_FORWARD_DELAY')
            elif ret['kind'] == 'vrf':
                ret['vrf_table'] = linux_net.get_attr(ifla_data,
                                                      'IFLA_VRF_TABLE')

        retval[device['index']] = ret

    for device in retval.values():
        if device.get('parent_index'):
            parent_device = retval.get(device['parent_index'])
            if parent_device:
                device['parent_name'] = parent_device['name']
        elif device.get('vxlan_link_index'):
            device['vxlan_link_name'] = (
                retval[device['vxlan_link_index']]['name'])

    return retval


class _LinuxNetTestCase(base_functional.BaseFunctionalTestCase):

    def setUp(self):
        super().setUp()
        self.dev_name = uuidutils.generate_uuid()[:15]
        self.dev_name2 = uuidutils.generate_uuid()[:15]
        self.addCleanup(self._delete_interface)

    def _delete_interface(self):
        def delete_device(device_name):
            try:
                linux_net.delete_interface(device_name)
            except Exception:
                pass

        if self._get_device(self.dev_name):
            delete_device(self.dev_name)
        if self._get_device(self.dev_name2):
            delete_device(self.dev_name2)

    def _get_device(self, device_name):
        devices = get_devices_info()
        for device in devices.values():
            if device['name'] == device_name:
                return device

    def _assert_state(self, device_name, state):
        device = self._get_device(device_name)
        return state == device['state']

    def _check_status(self, device_name):
        fn = functools.partial(self._assert_state, device_name,
                               constants.LINK_DOWN)
        test_utils.wait_until_true(fn, timeout=5)
        set_up(device_name)
        fn = functools.partial(self._assert_state, device_name,
                               constants.LINK_UP)
        test_utils.wait_until_true(fn, timeout=5)


class IpLinkTestCase(_LinuxNetTestCase):

    def test_create_interface_dummy(self):
        linux_net.create_interface(self.dev_name, 'dummy')
        device = self._get_device(self.dev_name)
        self.assertEqual('dummy', device['kind'])
        self._check_status(self.dev_name)

    def test_create_interface_vlan(self):
        vlan_id = random.randint(2, 4094)
        linux_net.create_interface(self.dev_name, 'dummy')
        linux_net.create_interface(self.dev_name2, 'vlan',
                                   physical_interface=self.dev_name,
                                   vlan_id=vlan_id)
        device = self._get_device(self.dev_name2)
        self.assertEqual('vlan', device['kind'])
        self.assertEqual(vlan_id, device['vlan_id'])
        self._check_status(self.dev_name)

    def test_create_interface_vxlan(self):
        vxlan_id = random.randint(2, 4094)
        vxlan_port = random.randint(10000, 65534)
        vxlan_local = '1.2.3.4'
        linux_net.create_interface(self.dev_name, 'vxlan',
                                   vxlan_id=vxlan_id,
                                   vxlan_port=vxlan_port,
                                   vxlan_local=vxlan_local,
                                   vxlan_learning=False,
                                   state=constants.LINK_UP)
        device = self._get_device(self.dev_name)
        self.assertEqual('vxlan', device['kind'])
        self.assertEqual(vxlan_id, device['vxlan_id'])
        self.assertEqual(vxlan_port, device['vxlan_port'])
        self.assertEqual(vxlan_local, device['vxlan_local'])
        self.assertEqual(constants.LINK_UP, device['state'])
        self.assertFalse(device['vxlan_learning'])

    def test_create_interface_veth(self):
        linux_net.create_interface(self.dev_name, 'veth', peer=self.dev_name2)
        device = self._get_device(self.dev_name)
        self.assertEqual('veth', device['kind'])
        self.assertEqual(self.dev_name2, device['parent_name'])
        device = self._get_device(self.dev_name2)
        self.assertEqual('veth', device['kind'])
        self.assertEqual(self.dev_name, device['parent_name'])
        self._check_status(self.dev_name)
        self._check_status(self.dev_name2)

    def test_create_interface_bridge(self):
        linux_net.create_interface(self.dev_name, 'bridge', br_stp_state=0)
        device = self._get_device(self.dev_name)
        self.assertEqual('bridge', device['kind'])
        self.assertEqual(0, device['stp'])
        self._check_status(self.dev_name)

    def test_create_interface_vrf(self):
        vrf_table = random.randint(10, 2000)
        linux_net.create_interface(self.dev_name, 'vrf', vrf_table=vrf_table)
        device = self._get_device(self.dev_name)
        self.assertEqual('vrf', device['kind'])
        self.assertEqual(vrf_table, device['vrf_table'])
        self._check_status(self.dev_name)

    def _check_device_master_vrf(self, device, master=None):
        device_info = self._get_device(device)
        if not master:
            self.assertIsNone(device_info['master'])
            self.assertIsNone(device_info['slave_kind'])
        else:
            master_info = self._get_device(master)
            self.assertEqual(master_info['index'], device_info['master'])
            self.assertEqual('vrf', device_info['slave_kind'])

    def test_set_master_for_device_bridge(self):
        vrf_table = random.randint(10, 2000)
        linux_net.create_interface(self.dev_name, 'vrf', vrf_table=vrf_table)
        linux_net.create_interface(self.dev_name2, 'bridge', br_stp_state=0)
        self._check_device_master_vrf(self.dev_name2)
        linux_net.set_master_for_device(self.dev_name2, self.dev_name)
        self._check_device_master_vrf(self.dev_name2, master=self.dev_name)

    def test_set_master_for_device_dummy(self):
        vrf_table = random.randint(10, 2000)
        linux_net.create_interface(self.dev_name, 'vrf', vrf_table=vrf_table)
        linux_net.create_interface(self.dev_name2, 'dummy')
        self._check_device_master_vrf(self.dev_name2)
        linux_net.set_master_for_device(self.dev_name2, self.dev_name)
        self._check_device_master_vrf(self.dev_name2, master=self.dev_name)

    def test_set_master_for_device_vlan(self):
        vrf_table = random.randint(10, 2000)
        linux_net.create_interface(self.dev_name, 'vrf', vrf_table=vrf_table)
        vlan_id = random.randint(2, 4094)
        dev_name3 = uuidutils.generate_uuid()[:15]
        linux_net.create_interface(self.dev_name2, 'dummy')
        linux_net.create_interface(dev_name3, 'vlan',
                                   physical_interface=self.dev_name2,
                                   vlan_id=vlan_id)
        self._check_device_master_vrf(dev_name3)
        linux_net.set_master_for_device(dev_name3, self.dev_name)
        self._check_device_master_vrf(dev_name3, master=self.dev_name)

    def test_set_master_for_device_veth(self):
        vrf_table = random.randint(10, 2000)
        linux_net.create_interface(self.dev_name, 'vrf', vrf_table=vrf_table)
        dev_name3 = uuidutils.generate_uuid()[:15]
        linux_net.create_interface(self.dev_name2, 'veth', peer=dev_name3)
        self._check_device_master_vrf(self.dev_name2)
        linux_net.set_master_for_device(self.dev_name2, self.dev_name)
        self._check_device_master_vrf(self.dev_name2, master=self.dev_name)

    def test_ensure_vlan_device_for_network(self):
        self.dev_name = uuidutils.generate_uuid()[:8]
        linux_net.create_interface(self.dev_name, 'dummy')
        linux_net.set_device_state(self.dev_name, constants.LINK_UP)
        vlan_id = random.randint(2, 4094)

        # Ensure the method call is idempotent.
        for _ in range(2):
            linux_net.ensure_vlan_device_for_network(self.dev_name, vlan_id)
            fn = functools.partial(self._assert_state, self.dev_name,
                                   constants.LINK_UP)
            test_utils.wait_until_true(fn, timeout=5)

    def test_ensure_vrf(self):
        vrf_table = random.randint(10, 2000)
        # Ensure the method call is idempotent.
        for _ in range(2):
            linux_net.ensure_vrf(self.dev_name, vrf_table)
            fn = functools.partial(self._assert_state, self.dev_name,
                                   constants.LINK_UP)
            test_utils.wait_until_true(fn, timeout=5)

    def test_ensure_bridge(self):
        # Ensure the method call is idempotent.
        for _ in range(2):
            linux_net.ensure_bridge(self.dev_name)
            fn = functools.partial(self._assert_state, self.dev_name,
                                   constants.LINK_UP)
            test_utils.wait_until_true(fn, timeout=5)

    def test_ensure_vxlan(self):
        vxlan_id = random.randint(2, 4094)
        vxlan_port = random.randint(10000, 65534)
        vxlan_local = '1.2.3.4'
        # Ensure the method call is idempotent.
        for _ in range(2):
            linux_net.ensure_vxlan(self.dev_name, vxlan_id, vxlan_local,
                                   vxlan_port)
            fn = functools.partial(self._assert_state, self.dev_name,
                                   constants.LINK_UP)
            test_utils.wait_until_true(fn, timeout=5)

    def test_ensure_veth(self):
        # Ensure the method call is idempotent.
        for _ in range(2):
            linux_net.ensure_veth(self.dev_name, self.dev_name2)
            fn = functools.partial(self._assert_state, self.dev_name,
                                   constants.LINK_UP)
            test_utils.wait_until_true(fn, timeout=5)

    def test_ensure_dummy(self):
        for _ in range(2):
            linux_net.ensure_dummy_device(self.dev_name)
            fn = functools.partial(self._assert_state, self.dev_name,
                                   constants.LINK_UP)
            test_utils.wait_until_true(fn, timeout=5)

    def test_get_bridge_vlan_devices(self):
        vlan_id = random.randint(2, 4094)
        linux_net.create_interface(self.dev_name, 'dummy')
        linux_net.create_interface(self.dev_name2, 'vlan',
                                   physical_interface=self.dev_name,
                                   vlan_id=vlan_id)

        vlan_devices = linux_net.get_bridge_vlans(self.dev_name)
        self.assertEqual(vlan_devices[0], vlan_id)


class IpAddressTestCase(_LinuxNetTestCase):

    def test_add_and_delete_ip_address(self):
        def check_ip_address(ip_address, device_name, present=True):
            ip_addresses = get_ip_addresses(self.dev_name)
            if l_net.get_ip_version(ip_address) == constants.IP_VERSION_6:
                address = '{}/128'.format(ip_address)
            else:
                address = '{}/32'.format(ip_address)
            for _ip in ip_addresses:
                if _ip['cidr'] == address:
                    if present:
                        return
                    else:
                        self.fail('IP address %s present in device %s' %
                                  (ip_address, device_name))

            if present:
                self.fail('IP address %s not found in device %s' %
                          (ip_address, device_name))

        ip_addresses = ('240.0.0.1', 'fd00::1')
        linux_net.create_interface(self.dev_name, 'dummy')
        for ip_address in ip_addresses:
            linux_net.add_ip_address(ip_address, self.dev_name)
            check_ip_address(ip_address, self.dev_name)
            # ensure nothing breaks if same IP gets added,
            # it should raise exception that is handled in the utils
            self.assertRaises(agent_exc.IpAddressAlreadyExists,
                              linux_net.add_ip_address, ip_address,
                              self.dev_name)

        for ip_address in ip_addresses:
            linux_net.delete_ip_address(ip_address, self.dev_name)
            check_ip_address(ip_address, self.dev_name, present=False)
            # ensure removing a missing IP is ok
            linux_net.delete_ip_address(ip_address, self.dev_name)

    def test_add_ip_address_no_device(self):
        self.assertRaises(agent_exc.NetworkInterfaceNotFound,
                          linux_net.add_ip_address, '240.0.0.1', self.dev_name)

    def test_delete_ip_address_no_device(self):
        self.assertRaises(agent_exc.NetworkInterfaceNotFound,
                          linux_net.delete_ip_address, '240.0.0.1',
                          self.dev_name)

    def test_delete_ip_address_no_ip_on_device(self):
        linux_net.create_interface(self.dev_name, 'dummy')
        # No exception is raised.
        linux_net.delete_ip_address('192.168.0.1', self.dev_name)


class IpRouteTestCase(_LinuxNetTestCase):

    def setUp(self):
        super().setUp()
        linux_net.create_interface(self.dev_name, 'dummy',
                                   state=constants.LINK_UP)
        self.device = self._get_device(self.dev_name)

    def _check_routes(self, cidrs, device_name, table=None, scope='link',
                      proto='static', route_present=True):
        table = table or iproute_linux.DEFAULT_TABLE
        cidr = None
        for cidr in cidrs:
            ip_version = l_net.get_ip_version(cidr)
            if ip_version == n_const.IP_VERSION_6:
                scope = 0
            if isinstance(scope, int):
                scope = linux_net.get_scope_name(scope)
            routes = linux_net.list_ip_routes(ip_version, device=device_name)
            for route in routes:
                ip = linux_net.get_attr(route, 'RTA_DST')
                mask = route['dst_len']
                if not (ip == str(netaddr.IPNetwork(cidr).ip) and
                        mask == netaddr.IPNetwork(cidr).cidr.prefixlen):
                    continue
                self.assertEqual(table, route['table'])
                self.assertEqual(
                    common_utils.IP_VERSION_FAMILY_MAP[ip_version],
                    route['family'])
                ret_scope = linux_net.get_scope_name(route['scope'])
                self.assertEqual(scope, ret_scope)
                self.assertEqual(rtnl.rt_proto[proto], route['proto'])
                break
            else:
                if route_present:
                    self.fail('CIDR %s not found in the list of routes' % cidr)
                else:
                    return
        if not route_present:
            self.fail('CIDR %s found in the list of routes' % cidr)

    def _add_route_device_and_check(self, cidrs, table=None, scope='link',
                                    proto='static'):
        for cidr in cidrs:
            ip_version = l_net.get_ip_version(cidr)
            family = common_utils.IP_VERSION_FAMILY_MAP[ip_version]
            route = {'dst': cidr,
                     'oif': self.device['index'],
                     'table': table,
                     'family': family,
                     'scope': scope,
                     'proto': proto}
            linux_net.route_create(route)
            # recreate route to ensure it does not break anything
            linux_net.route_create(route)
        self._check_routes(cidrs, self.dev_name, table=table, scope=scope,
                           proto=proto)
        for cidr in cidrs:
            ip_version = l_net.get_ip_version(cidr)
            family = common_utils.IP_VERSION_FAMILY_MAP[ip_version]
            route = {'dst': cidr,
                     'oif': self.device['index'],
                     'table': table,
                     'family': family,
                     'scope': scope,
                     'proto': proto}
            linux_net.route_delete(route)
            # redelete route to ensure it does not break anything
            linux_net.route_delete(route)
        self._check_routes(cidrs, self.dev_name, table=table, scope=scope,
                           proto=proto, route_present=False)

    def test_add_route_device(self):
        cidrs = ['192.168.1.0/24', '2001:db1::/64']
        self._add_route_device_and_check(cidrs=cidrs, table=None)

    def test_add_route_device_table(self):
        cidrs = ['192.168.2.0/24', '2001:db2::/64']
        self._add_route_device_and_check(cidrs=cidrs, table=100)

    def test_add_route_device_scope_site(self):
        cidrs = ['192.168.3.0/24', '2003:db3::/64']
        self._add_route_device_and_check(cidrs=cidrs, scope='site')

    def test_add_route_device_scope_host(self):
        cidrs = ['192.168.4.0/24', '2003:db4::/64']
        self._add_route_device_and_check(cidrs=cidrs, scope='host')

    def test_add_route_device_proto_static(self):
        cidrs = ['192.168.5.0/24', '2003:db5::/64']
        self._add_route_device_and_check(cidrs=cidrs, proto='static')

    def test_add_route_device_proto_redirect(self):
        cidrs = ['192.168.6.0/24', '2003:db6::/64']
        self._add_route_device_and_check(cidrs=cidrs, proto='redirect')

    def test_add_route_device_proto_kernel(self):
        cidrs = ['192.168.7.0/24', '2003:db7::/64']
        self._add_route_device_and_check(cidrs=cidrs, proto='kernel')

    def test_add_route_device_proto_boot(self):
        cidrs = ['192.168.8.0/24', '2003:db8::/64']
        self._add_route_device_and_check(cidrs=cidrs, proto='boot')

    def test_add_unreachable_route(self):
        vrf_table = random.randint(10, 2000)
        linux_net.create_interface(self.dev_name2, 'vrf', vrf_table=vrf_table)
        linux_net.add_unreachable_route(self.dev_name2)
        for ip_version in (n_const.IP_VERSION_4, n_const.IP_VERSION_6):
            routes = linux_net.list_ip_routes(ip_version, table=vrf_table)
            self.assertEqual(1, len(routes))
            self.assertEqual(rtnl.rt_proto['boot'], routes[0]['proto'])
            self.assertEqual(rtnl.rtypes['RTN_UNREACHABLE'], routes[0]['type'])
            self.assertEqual(4278198272,
                             linux_net.get_attr(routes[0], 'RTA_PRIORITY'))


class IpRuleTestCase(_LinuxNetTestCase):

    def _test_add_and_delete_ip_rule(self, ip_version, cidrs):
        table = random.randint(10, 250)
        rules_added = []
        for cidr in cidrs:
            _ip = netaddr.IPNetwork(cidr)
            rule = {'dst': str(_ip.ip),
                    'dst_len': _ip.netmask.netmask_bits(),
                    'table': table,
                    'family': common_utils.IP_VERSION_FAMILY_MAP[ip_version]}
            rules_added.append(rule)
            linux_net.rule_create(rule)
            # recreate the last rule, to ensure recreation does not fail
            linux_net.rule_create(rule)
        rules = linux_net.list_ip_rules(ip_version, table=table)
        self.assertEqual(len(cidrs), len(rules))
        for idx, rule in enumerate(rules):
            _ip = netaddr.IPNetwork(cidrs[idx])
            self.assertEqual(str(_ip.ip), linux_net.get_attr(rule, 'FRA_DST'))
            self.assertEqual(_ip.netmask.netmask_bits(), rule['dst_len'])

        for rule in rules_added:
            linux_net.rule_delete(rule)
            # remove again the last rule to ensure it does not fail
            linux_net.rule_delete(rule)
        rules = linux_net.list_ip_rules(ip_version, table=table)
        self.assertEqual(0, len(rules))

    def test_add_and_delete_ip_rule_v4(self):
        cidrs = ['192.168.0.0/24', '172.90.0.0/16', '10.0.0.0/8']
        self._test_add_and_delete_ip_rule(n_const.IP_VERSION_4, cidrs)

    def test_add_and_delete_ip_rule_v6(self):
        cidrs = ['2001:db8::/64', 'fe80::/10']
        self._test_add_and_delete_ip_rule(n_const.IP_VERSION_6, cidrs)


class IpNeighTestCase(_LinuxNetTestCase):

    def setUp(self):
        super().setUp()
        linux_net.create_interface(self.dev_name, 'dummy',
                                   state=constants.LINK_UP)
        self.device = self._get_device(self.dev_name)

    def test_add_and_delete_neigh(self):
        # Initial check, nothing in the ARP table.
        neigh4 = linux_net.get_neigh_entries(self.dev_name,
                                             constants.IP_VERSION_4)
        neigh6 = linux_net.get_neigh_entries(self.dev_name,
                                             constants.IP_VERSION_6)
        self.assertEqual([], neigh4)
        self.assertEqual([], neigh6)

        # Add a set of IP/MAC addresses.
        ip_and_mac = [('10.0.0.1', 'ca:fe:ca:fe:00:01'),
                      ('10.0.0.2', 'ca:fe:ca:fe:00:02'),
                      ('2001:db8::3', 'ca:fe:ca:fe:00:03'),
                      ('2001:db8::4', 'ca:fe:ca:fe:00:04')]
        for ip, mac in ip_and_mac:
            linux_net.add_ip_nei(ip, mac, self.dev_name)
        neigh4 = linux_net.get_neigh_entries(self.dev_name,
                                             constants.IP_VERSION_4)
        neigh6 = linux_net.get_neigh_entries(self.dev_name,
                                             constants.IP_VERSION_6)
        self.assertEqual(2, len(neigh4))
        self.assertEqual(2, len(neigh6))
        for neigh in neigh4 + neigh6:
            for ip, mac in ip_and_mac:
                if ip == neigh['dst'] and mac == neigh['lladdr']:
                    break
            else:
                self.fail('IP/MAC %s/%s is not present in the ip-neigh table' %
                          (neigh['dst'], neigh['lladdr']))

        # Delete the entries.
        for ip, mac in ip_and_mac:
            linux_net.del_ip_nei(ip, mac, self.dev_name)
        neigh4 = linux_net.get_neigh_entries(self.dev_name,
                                             constants.IP_VERSION_4)
        neigh6 = linux_net.get_neigh_entries(self.dev_name,
                                             constants.IP_VERSION_6)
        self.assertEqual(0, len(neigh4))
        self.assertEqual(0, len(neigh6))



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_concurrency import processutils
from oslo_log import log as logging

from ovn_bgp_agent import constants
import ovn_bgp_agent.privileged.vtysh

LOG = logging.getLogger(__name__)


@ovn_bgp_agent.privileged.vtysh_cmd.entrypoint
def run_vtysh_config(frr_config_file):
    full_args = ['/usr/bin/vtysh', '--vty_socket', constants.FRR_SOCKET_PATH,
                 '-f', frr_config_file]
    try:
        return processutils.execute(*full_args)
    except Exception as e:
        LOG.exception("Unable to execute vtysh with %s. Exception: %s",
                      full_args, e)
        raise


@ovn_bgp_agent.privileged.vtysh_cmd.entrypoint
def run_vtysh_command(command):
    full_args = ['/usr/bin/vtysh', '--vty_socket', constants.FRR_SOCKET_PATH,
                 '-c', command]
    try:
        return processutils.execute(*full_args)[0]
    except Exception as e:
        LOG.exception("Unable to execute vtysh with %s. Exception: %s",
                      full_args, e)
        raise



# Copyright 2024 team.blue/nl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from unittest import mock

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import evpn
from ovn_bgp_agent import exceptions
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.tests import utils


CONF = cfg.CONF


class TestEVPN(test_base.TestCase):

    def setUp(self):
        super(TestEVPN, self).setUp()
        CONF.set_override('evpn_local_ip', '127.0.0.1')

        self.mock_ovs = mock.patch.object(evpn, 'ovs').start()
        self.mock_ovs.get_ovs_patch_port_ofport.return_value = 12

        self.mock_frr = mock.patch.object(evpn, 'frr').start()
        self.mock_linux_net = mock.patch.object(evpn, 'linux_net').start()

        self.fake_mac = fake_mac = 'fe:12:34:56:89:90'
        self.mock_linux_net.get_interface_address.return_value = fake_mac

        self._bridge_args = {
            'ovs_bridge': 'br-ex',
            'vni': 100,
            'evpn_opts': {},
            'mode': constants.OVN_EVPN_TYPE_L3,
            'ovs_flows': {'br-ex': {}},
        }
        self.vrf_name = 'vrf-100'
        self.vxlan_name = 'vxlan-100'
        self.bridge_name = 'br-100'

        self.veth_vrf = '_to_be_set_by__create_bridge_and_vlan'
        self.veth_ovs = '_to_be_set_by__create_bridge_and_vlan'

        # evpn.local_bridges = {}
    def _reset_evpn_local_bridges(self):
        evpn.local_bridges = {}

    def _create_bridge(self, **override_args) -> evpn.EvpnBridge:
        self.addCleanup(self._reset_evpn_local_bridges)
        args = dict(**self._bridge_args)
        args.update(**override_args)
        return evpn.setup(**args)

    def _create_bridge_and_vlan(
        self, vlan_tag=4094, **bridge_args
    ) -> 'tuple[object, evpn.EvpnBridge, evpn.VlanDev]':
        port = utils.create_row(
            name='fake-port-name',
            tag=[vlan_tag]
        )

        uuid = str(port.uuid)[0:11]
        self.veth_vrf = constants.OVN_EVPN_VETH_VRF_UUID_PREFIX + uuid
        self.veth_ovs = constants.OVN_EVPN_VETH_OVS_UUID_PREFIX + uuid

        evpn_bridge = self._create_bridge(**bridge_args)
        return port, evpn_bridge, evpn_bridge.connect_vlan(port)

    def test_setup_no_ip(self):
        CONF.set_override('evpn_local_ip', None)
        self.assertRaises(exceptions.ConfOptionRequired, self._create_bridge)

    def test_setup(self):
        evpn_bridge = self._create_bridge()

        self.assertIsInstance(evpn_bridge, evpn.EvpnBridge)
        self.assertEqual(evpn_bridge.ovs_bridge, 'br-ex')
        self.assertEqual(evpn_bridge.vni, 100)
        self.assertEqual(evpn_bridge.vrf_name, self.vrf_name)
        self.assertEqual(evpn_bridge.bridge_name, self.bridge_name)
        self.assertEqual(evpn_bridge.vxlan_name, self.vxlan_name)

        other_bridge = self._create_bridge()
        self.assertEqual(evpn_bridge, other_bridge)

    def test_lookup_vlan_int(self):
        port = utils.create_row(
            tag=[4094]
        )
        evpn_bridge = self._create_bridge()
        evpn_bridge.connect_vlan(port)

        bridge = evpn.lookup(self._bridge_args['ovs_bridge'], 4094)
        self.assertIsInstance(bridge, evpn.EvpnBridge)

    def test_lookup_vlan_str(self):
        port = utils.create_row(
            tag=[4094]
        )
        evpn_bridge = self._create_bridge()
        evpn_bridge.connect_vlan(port)

        bridge = evpn.lookup(self._bridge_args['ovs_bridge'], '4094')
        self.assertIsInstance(bridge, evpn.EvpnBridge)

    def test_lookup_vlan_None(self):
        port = utils.create_row(
            tag=[]
        )
        evpn_bridge = self._create_bridge()
        evpn_bridge.connect_vlan(port)

        bridge = evpn.lookup(self._bridge_args['ovs_bridge'], None)
        self.assertIsInstance(bridge, evpn.EvpnBridge)

    def test_lookup_vlan_unknown(self):
        self._create_bridge_and_vlan(vlan_tag=123)
        self._create_bridge_and_vlan(vlan_tag=123, vni=123, ovs_bridge='foo')
        self._create_bridge_and_vlan(vlan_tag=4094, vni=123, ovs_bridge='foo')
        self.assertRaises(KeyError, evpn.lookup,
                          self._bridge_args['ovs_bridge'], '4094')

    def test_evpnbridge_setup_l3(self):
        bridge = self._create_bridge()
        bridge.setup()

        self.assertTrue(bridge._setup_done)

        # Create pointer for shorter lines.
        linux_net = self.mock_linux_net

        linux_net.ensure_bridge.assert_called_once_with(self.bridge_name)
        linux_net.ensure_vxlan.assert_called_once_with(self.vxlan_name, 100,
                                                       '127.0.0.1', 4789)
        linux_net.set_master_for_device.assert_has_calls([
            mock.call(self.vxlan_name, self.bridge_name),
            mock.call(self.bridge_name, self.vrf_name),
        ])
        linux_net.disable_learning_vxlan_intf.assert_called_once_with(
            self.vxlan_name)
        linux_net.ensure_vrf.assert_called_once_with(self.vrf_name, 100)

        frr = self.mock_frr
        frr.vrf_reconfigure.assert_called_once_with(mock.ANY, 'add-vrf')

    def test_evpnbridge_setup_l2(self):
        bridge = self._create_bridge(mode=constants.OVN_EVPN_TYPE_L2)
        bridge.setup()

        self.assertTrue(bridge._setup_done)

        # Create pointer for shorter lines.
        linux_net = self.mock_linux_net

        linux_net.ensure_bridge.assert_called_once_with(self.bridge_name)
        linux_net.ensure_vxlan.assert_called_once_with(self.vxlan_name, 100,
                                                       '127.0.0.1', 4789)
        linux_net.set_master_for_device.assert_has_calls([
            mock.call(self.vxlan_name, self.bridge_name),
        ])
        linux_net.disable_learning_vxlan_intf.assert_called_once_with(
            self.vxlan_name)
        linux_net.ensure_vrf.assert_not_called()

        frr = self.mock_frr
        frr.vrf_reconfigure.assert_called_once_with(mock.ANY, 'add-vrf')

    def test_evpnbridge_setup_done(self):
        bridge = self._create_bridge()
        bridge._setup_done = True
        bridge.setup()

        self.mock_linux_net.ensure_bridge.assert_not_called()

    def test_evpnbridge_eval_disconnect(self):
        _, bridge, evpn_vlan = self._create_bridge_and_vlan()

        bridge_disconnect = mock.patch.object(bridge, 'disconnect').start()

        bridge._eval_disconnect()
        bridge_disconnect.assert_not_called()

        bridge._setup_done = True
        evpn_vlan._setup_done = True
        bridge._eval_disconnect()

        bridge_disconnect.assert_not_called()

        evpn_vlan._setup_done = False

        bridge._eval_disconnect()
        bridge_disconnect.assert_called_once()

    def test_evpnbridge_disconnect(self):
        bridge = self._create_bridge()
        bridge._setup_done = True
        bridge.disconnect()

        calls = [mock.call('br-100'),
                 mock.call('vxlan-100'),
                 mock.call('vrf-100')]
        self.mock_linux_net.delete_device.assert_has_calls(calls)
        self.mock_frr.vrf_reconfigure.assert_called_once_with(mock.ANY,
                                                              action='del-vrf')

        self.assertFalse(bridge._setup_done)

    def test_evpnbridge_disconnect_keep_vrf(self):
        bridge = self._create_bridge()
        bridge._setup_done = True
        CONF.set_override('delete_vrf_on_disconnect', False)
        bridge.disconnect()

        calls = [mock.call('br-100'),
                 mock.call('vxlan-100')]
        self.mock_linux_net.delete_device.assert_has_calls(calls)
        self.mock_frr.vrf_reconfigure.assert_not_called()

        self.assertFalse(bridge._setup_done)

    def test_evpnbridge_connect_vlan_again(self):
        port, bridge, evpn_vlan = self._create_bridge_and_vlan()

        vlan = bridge.connect_vlan(port)
        self.assertEqual(vlan, evpn_vlan)

    def test_evpnbridge_get_vlan(self):
        _, bridge, evpn_vlan = self._create_bridge_and_vlan()
        self.assertEqual(bridge.get_vlan(4094), evpn_vlan)

    def test_evpnbridge_vlan_lladdr_property_calls_setup(self):
        _, bridge, evpn_vlan = self._create_bridge_and_vlan()

        evpn_vlan_setup = mock.patch.object(evpn_vlan, 'setup').start()
        self.assertEqual(evpn_vlan.lladdr, self.fake_mac)
        evpn_vlan_setup.assert_called_once()

    def test_evpnbridge_vlan_setup_l2(self):
        vlan_tag = 4094
        vlan_tag_str = '4094'
        _, evpn_bridge, vlan_dev = self._create_bridge_and_vlan(vlan_tag,
                                                                mode='l2')

        evpn_setup = mock.patch.object(evpn_bridge, 'setup').start()

        vlan_dev.setup()

        evpn_setup.assert_called_once()

        linux_net = self.mock_linux_net
        linux_net.ensure_veth.assert_called_once_with(self.veth_vrf,
                                                      self.veth_ovs)
        self.mock_ovs.add_device_to_ovs_bridge(self.veth_ovs, 'br-ex',
                                               vlan_tag=vlan_tag_str)
        linux_net.set_master_for_device.assert_called_once_with(self.veth_vrf,
                                                                'br-100')

        linux_net.ensure_arp_ndp_enabled_for_bridge.assert_not_called()
        linux_net.enable_routing_for_interfaces.assert_not_called()
        linux_net.enable_proxy_arp.assert_not_called()
        linux_net.enable_proxy_ndp.assert_not_called()

        self.mock_ovs.ensure_mac_tweak_flows.assert_not_called()
        self.mock_ovs.remove_extra_ovs_flows.assert_not_called()

        self.assertTrue(vlan_dev._veth_created)
        self.assertTrue(vlan_dev._setup_done)

    def test_evpnbridge_vlan_setup_l3(self, custom_ips=[]):
        vlan_tag = 4094
        vlan_tag_str = '4094'
        _, evpn_bridge, vlan_dev = self._create_bridge_and_vlan(vlan_tag)

        evpn_setup = mock.patch.object(evpn_bridge, 'setup').start()

        if custom_ips:
            vlan_dev.add_ips(list(custom_ips))

        vlan_dev.setup()

        evpn_setup.assert_called_once()

        linux_net = self.mock_linux_net
        linux_net.ensure_veth.assert_called_once_with(self.veth_vrf,
                                                      self.veth_ovs)
        self.mock_ovs.add_device_to_ovs_bridge(self.veth_ovs, 'br-ex',
                                               vlan_tag=vlan_tag_str)
        linux_net.set_master_for_device.assert_called_once_with(self.veth_vrf,
                                                                self.vrf_name)
        linux_net.ensure_arp_ndp_enabled_for_bridge.assert_called_once_with(
            self.veth_vrf, offset=vlan_tag, vlan_tag=vlan_tag_str)

        linux_net.enable_routing_for_interfaces.assert_called_once_with(
            self.veth_vrf, 'br-100')

        linux_net.enable_proxy_arp.assert_called_once_with(self.veth_vrf)
        linux_net.enable_proxy_ndp.assert_called_once_with(self.veth_vrf)

        self.mock_ovs.ensure_mac_tweak_flows.assert_called_once_with(
            'br-ex', self.fake_mac, [12], constants.OVS_RULE_COOKIE)
        self.mock_ovs.remove_extra_ovs_flows.assert_called_once_with(
            mock.ANY, 'br-ex', constants.OVS_RULE_COOKIE)

        self.assertTrue(vlan_dev._veth_created)
        self.assertTrue(vlan_dev._setup_done)

        if custom_ips:
            linux_net.add_ips_to_dev.assert_has_calls([
                mock.call(self.veth_vrf, ips=custom_ips),
            ])

        linux_net.ensure_anycast_mac_for_interface.assert_called_once_with(
            self.veth_vrf, offset=6557694)

    def test_evpnbridge_vlan_setup_l3_with_custom_ips(self):
        self.test_evpnbridge_vlan_setup_l3(custom_ips=['10.10.10.1'])

    def test_evpnbridge_vlan_setup_l3_failed_ovs_call(self):
        vlan_tag = 4094
        _, evpn_bridge, vlan_dev = self._create_bridge_and_vlan(vlan_tag)

        mock.patch.object(evpn_bridge, 'setup').start()

        self.mock_ovs.get_ovs_patch_port_ofport.side_effect = KeyError

        vlan_dev.setup()
        self.assertTrue(vlan_dev._veth_created)
        self.assertFalse(vlan_dev._setup_done)

    def test_evpnbridge_vlan__eval_disconnect(self):
        _, _, vlan_dev = self._create_bridge_and_vlan()

        vlan_dev_disconnect = mock.patch.object(vlan_dev, 'disconnect').start()

        vlan_dev._setup_done = True

        vlan_dev._agent_routing_tables_routes = ['entry']
        vlan_dev._eval_disconnect()
        vlan_dev_disconnect.assert_not_called()

        vlan_dev._agent_routing_tables_routes = []
        vlan_dev._eval_disconnect()
        vlan_dev_disconnect.assert_called_once()

    def test_evpnbridge_vlan__eval_disconnect_not_setup_yet(self):
        _, _, vlan_dev = self._create_bridge_and_vlan()

        vlan_dev_disconnect = mock.patch.object(vlan_dev, 'disconnect').start()

        vlan_dev._agent_routing_tables_routes = []
        vlan_dev._eval_disconnect()

        vlan_dev_disconnect.assert_not_called()

    def test_evpnbridge_vlan_disconnect(self):
        _, evpn_bridge, vlan_dev = self._create_bridge_and_vlan()

        evpn_bridge__eval_disconnect = mock.patch.object(
            evpn_bridge, '_eval_disconnect').start()

        vlan_dev.disconnect()

        evpn_bridge__eval_disconnect.assert_called_once()

        self.mock_ovs.del_device_from_ovs_bridge.assert_called_once_with(
            self.veth_ovs, 'br-ex')
        self.mock_linux_net.delete_device.assert_called_once_with(
            self.veth_vrf)

        self.assertFalse(vlan_dev._veth_created)
        self.assertFalse(vlan_dev._setup_done)

    def test_evpnbridge_vlan_teardown(self):
        _, evpn_bridge, vlan_dev = self._create_bridge_and_vlan()

        vlan_dev_disconnect = mock.patch.object(
            vlan_dev, 'disconnect').start()

        vlan_dev.teardown()

        vlan_dev_disconnect.assert_called_once()
        self.assertNotIn(vlan_dev.vlan_tag, evpn_bridge.vlans)

    def test_evpnbridge_vlan_process_dhcp_opts(self):
        _, _, vlan_dev = self._create_bridge_and_vlan()
        dhcp_opts = [
            utils.create_row(cidr='10.10.10.0/24',
                             options={'router': '10.10.10.1'}),
            utils.create_row(cidr='fe00::/64', options={}),
        ]
        vlan_dev._setup_done = True
        vlan_dev.process_dhcp_opts(dhcp_opts)

        ips = {'10.10.10.1'}
        self.assertSetEqual(vlan_dev._custom_ips, ips)
        self.mock_linux_net.add_ips_to_dev.assert_called_once_with(
            self.veth_vrf, ips=list(ips))
        self.mock_frr.nd_reconfigure.assert_called_once_with(
            self.veth_vrf, 'fe00::/64', {})

    def test_evpnbridge_vlan_process_dhcp_opts_multiple_subnets(self):
        _, _, vlan_dev = self._create_bridge_and_vlan()
        dhcp_opts = [
            utils.create_row(cidr='10.10.10.0/24',
                             options={'router': '10.10.10.1'}),
            utils.create_row(cidr='10.10.20.0/24',
                             options={'router': '10.10.20.1'}),
        ]
        vlan_dev._setup_done = True
        vlan_dev.process_dhcp_opts(dhcp_opts)

        ips = {'10.10.10.1', '10.10.20.1'}
        self.assertSetEqual(vlan_dev._custom_ips, ips)

        calls = [
            mock.call(self.veth_vrf, ips=['10.10.10.1']),
            mock.call(self.veth_vrf, ips=['10.10.20.1']),
        ]
        self.mock_linux_net.add_ips_to_dev.assert_has_calls(calls)

    def test_evpnbridge_vlan_add_route(self, ip='10.10.10.10'):
        _, _, vlan_dev = self._create_bridge_and_vlan()
        vlan_dev._setup_done = True

        routing_tables_routes = {self.veth_vrf: []}
        addr = ip.split('/')[0]
        mask = None if '/' not in ip else ip.split('/')[1]
        mac = 'fe:12:34:56:89:12'
        via = None

        vlan_dev.add_route(routing_tables_routes, ip, mac, via)

        self.assertDictEqual(routing_tables_routes, {self.veth_vrf: [{
            'ip': '10.10.10.10',
            'mask': mask,
            'mac': 'fe:12:34:56:89:12',
            'via': None,
        }]})

        self.mock_linux_net.add_ip_route.assert_called_once_with(
            mock.ANY, addr, 100, self.veth_vrf, mask=mask, via=None)

        self.mock_linux_net.add_ip_nei.assert_called_once_with(
            addr, 'fe:12:34:56:89:12', self.veth_vrf)

    def test_evpnbridge_vlan_add_route_with_prefix(self):
        self.test_evpnbridge_vlan_add_route(ip='10.10.10.10/32')

    def test_evpnbridge_vlan_add_route_l2(self):
        _, _, vlan_dev = self._create_bridge_and_vlan(mode='l2')
        vlan_dev._setup_done = True

        routing_tables_routes = {self.veth_vrf: []}
        ip = '10.10.10.10/32'
        mac = 'fe:12:34:56:89:12'
        via = None

        vlan_dev.add_route(routing_tables_routes, ip, mac, via)

        self.mock_linux_net.add_ip_route.assert_not_called()

    def test_evpnbridge_vlan_del_route(self, ip='10.10.10.10'):
        _, _, vlan_dev = self._create_bridge_and_vlan()
        vlan_dev._setup_done = True

        addr = ip.split('/')[0]
        routing_tables_routes = {self.veth_vrf: [{
            'ip': addr,
            'mask': '32',
            'mac': 'fe:12:34:56:89:12',
            'via': '10.10.20.10',
        }, {
            'ip': '10.10.10.11',
            'mask': '32',
            'mac': 'fe:12:34:56:89:12',
            'via': '10.10.20.10',
        }]}
        mac = 'fe:12:34:56:89:12'
        vlan_dev__eval_disconnect = mock.patch.object(
            vlan_dev, '_eval_disconnect').start()

        vlan_dev.del_route(routing_tables_routes, ip, mac)

        self.assertDictEqual(routing_tables_routes, {self.veth_vrf: [{
            'ip': '10.10.10.11',
            'mask': '32',
            'mac': 'fe:12:34:56:89:12',
            'via': '10.10.20.10',
        }]})

        self.mock_linux_net.del_ip_route.assert_called_once_with(
            mock.ANY, addr, 100, self.veth_vrf, mask='32', via='10.10.20.10')

        self.mock_linux_net.del_ip_nei.assert_called_once_with(
            addr, 'fe:12:34:56:89:12', self.veth_vrf)

        vlan_dev__eval_disconnect.assert_called()

    def test_evpnbridge_vlan_del_route_with_prefix(self):
        self.test_evpnbridge_vlan_del_route('10.10.10.10/32')

    def test_evpnbridge_vlan_del_route_no_route_table(self):
        _, _, vlan_dev = self._create_bridge_and_vlan()
        vlan_dev._setup_done = True

        addr = '10.10.10.10'
        routing_tables_routes = {
            self.veth_vrf: [{
                'ip': '10.10.10.11',
                'mask': '32',
                'mac': 'fe:12:34:56:89:12',
                'via': '10.10.20.10',
            }]
        }
        mac = 'fe:12:34:56:89:12'
        vlan_dev__eval_disconnect = mock.patch.object(
            vlan_dev, '_eval_disconnect').start()

        vlan_dev.del_route(routing_tables_routes, addr, mac)

        self.assertDictEqual(routing_tables_routes, {self.veth_vrf: [{
            'ip': '10.10.10.11',
            'mask': '32',
            'mac': 'fe:12:34:56:89:12',
            'via': '10.10.20.10',
        }]})

        self.mock_linux_net.del_ip_route.assert_called_once_with(
            mock.ANY, addr, 100, self.veth_vrf, mask=None, via=None)

        self.mock_linux_net.del_ip_nei.assert_called_once_with(
            addr, 'fe:12:34:56:89:12', self.veth_vrf)

        vlan_dev__eval_disconnect.assert_called()

    def test_evpnbridge_vlan_del_route_l2(self):
        _, _, vlan_dev = self._create_bridge_and_vlan(mode='l2')
        vlan_dev._setup_done = True

        routing_tables_routes = {self.veth_vrf: [{
            'ip': '10.10.10.10',
            'mask': '32',
            'mac': 'fe:12:34:56:89:12',
            'via': '10.10.20.10',
        }]}
        ip = '10.10.10.10'
        mac = 'fe:12:34:56:89:12'

        vlan_dev.del_route(routing_tables_routes, ip, mac)

        self.assertDictEqual(routing_tables_routes, {self.veth_vrf: [{
            'ip': '10.10.10.10',
            'mask': '32',
            'mac': 'fe:12:34:56:89:12',
            'via': '10.10.20.10',
        }]})

        self.mock_linux_net.del_ip_route.assert_not_called()

    def test_evpnbridge_vlan_cleanup_excessive_routes(self):
        _, _, vlan_dev = self._create_bridge_and_vlan()
        vlan_dev._setup_done = True

        intf_idx = 1337
        self.mock_linux_net.get_interface_index.return_value = intf_idx

        routes = utils.create_linux_routes([{
            '_attrs': [
                ('RTA_DST', '198.51.100.0'), ('RTA_OIF', intf_idx),
                ('RTA_GATEWAY', '100.64.0.102')
            ],
            'dst_len': 28, 'type': 1,
        }, {
            'attrs': [('RTA_DST', '198.51.100.136'), ('RTA_OIF', intf_idx)],
            'dst_len': 32, 'type': 1,
        }, {
            'attrs': [('RTA_DST', '198.51.100.158'), ('RTA_OIF', intf_idx)],
            'dst_len': 32, 'type': 1,
        }])
        self.mock_linux_net._get_table_routes.return_value = routes

        routing_tables_routes = {self.veth_vrf: [{
            'ip': '198.51.100.0',
            'mask': '28',
            'mac': 'fe:12:34:56:89:12',
            'via': '100.64.0.102',
        }]}
        del_route = mock.patch.object(vlan_dev, 'del_route').start()

        vlan_dev.cleanup_excessive_routes(routing_tables_routes)

        calls = [
            mock.call(mock.ANY, '198.51.100.136'),
            mock.call(mock.ANY, '198.51.100.158'),
        ]
        del_route.assert_has_calls(calls, any_order=True)

    def test_evpnbridge_vlan_cleanup_excessive_routes_in_sync(self):
        _, _, vlan_dev = self._create_bridge_and_vlan()
        vlan_dev._setup_done = True

        intf_idx = 1337
        self.mock_linux_net.get_interface_index.return_value = intf_idx

        routes = utils.create_linux_routes([{
            '_attrs': [
                ('RTA_DST', '198.51.100.0'), ('RTA_OIF', intf_idx),
                ('RTA_GATEWAY', '100.64.0.102')
            ],
            'dst_len': 28, 'type': 1,
        }])
        self.mock_linux_net._get_table_routes.return_value = routes

        routing_tables_routes = {self.veth_vrf: [{
            'ip': '198.51.100.0',
            'mask': '28',
            'mac': 'fe:12:34:56:89:12',
            'via': '100.64.0.102',
        }]}
        del_route = mock.patch.object(vlan_dev, 'del_route').start()

        vlan_dev.cleanup_excessive_routes(routing_tables_routes)
        del_route.assert_not_called()

    def test_evpnbridge_vlan_cleanup_excessive_routes_not_setup_yet(self):
        _, _, vlan_dev = self._create_bridge_and_vlan()
        vlan_dev.cleanup_excessive_routes({})
        self.mock_linux_net._get_table_routes.assert_not_called()

    def test_evpn__find_route_info(self):
        result = evpn._find_route_info([
            {'ip': '198.51.100.136', 'mask': '32', 'mac': None, 'via': None},
            {'ip': '198.51.100.158', 'mask': '24', 'mac': None, 'via': None},
            {'ip': '127.0.0.1', 'mask': '8', 'mac': None, 'via': None},
            {'ip': '198.51.100.0', 'mask': '28', 'mac': None, 'via': None},
        ], '127.0.0.1')
        self.assertDictEqual(result, {'ip': '127.0.0.1', 'mask': '8',
                                      'mac': None, 'via': None})

    def test_evpn__find_route_info_not_found(self):
        result = evpn._find_route_info([], '127.0.0.1')
        self.assertDictEqual(result, {'ip': '127.0.0.1', 'mask': None,
                                      'mac': None, 'via': None})

    def test_evpn__ensure_list(self):
        self.assertListEqual(evpn._ensure_list(None), [])
        self.assertListEqual(evpn._ensure_list('aa'), ['aa'])
        self.assertListEqual(evpn._ensure_list(['aa']), ['aa'])
        self.assertSetEqual(evpn._ensure_list({'aa'}), {'aa'})
        self.assertTupleEqual(evpn._ensure_list(('a', 'b',)), ('a', 'b'))

    def test__offset_for_vni_and_vlan(self):
        vni = 100
        vlan = 100
        exp = int(('%x' % vni).zfill(6) + ('%x' % vlan).zfill(4), 16)
        self.assertEqual(exp, evpn._offset_for_vni_and_vlan(vni, vlan))

        vni = 16777214
        vlan = 4094
        exp = int(('%x' % vni).zfill(6) + ('%x' % vlan).zfill(4), 16)
        self.assertEqual(exp, evpn._offset_for_vni_and_vlan(vni, vlan))

        # Below value is 1 too much, so it is being modulo'd, and should be
        # calculated to 0
        vni = 16777215
        vlan = 4095
        exp = int(('%x' % vni).zfill(6) + ('%x' % vlan).zfill(4), 16)
        self.assertNotEqual(exp, evpn._offset_for_vni_and_vlan(vni, vlan))
        self.assertEqual(0, evpn._offset_for_vni_and_vlan(vni, vlan))



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import ipaddress
import os

import netaddr

from neutron_lib import constants as n_const
from oslo_concurrency import processutils
from oslo_log import log as logging
import pyroute2
from pyroute2 import iproute
from pyroute2 import netlink as pyroute_netlink
from pyroute2.netlink import exceptions as netlink_exceptions
from pyroute2.netlink import rtnl
from pyroute2.netlink.rtnl import ndmsg
import tenacity

import ovn_bgp_agent
from ovn_bgp_agent import constants
from ovn_bgp_agent import exceptions as agent_exc
from ovn_bgp_agent.utils import common as common_utils
from ovn_bgp_agent.utils import linux_net as l_net

LOG = logging.getLogger(__name__)

NUD_STATES = {state[1]: state[0] for state in ndmsg.states.items()}


def get_scope_name(scope):
    """Return the name of the scope or the scope number if the name is unknown.

    For backward compatibility (with "ip" tool) "global" scope is converted to
    "universe" before converting to number
    """
    scope = 'universe' if scope == 'global' else scope
    return rtnl.rt_scope.get(scope, scope)


def set_device_state(device, state):
    set_link_attribute(device, state=state)


@ovn_bgp_agent.privileged.default.entrypoint
def ensure_vrf(vrf_name, vrf_table):
    try:
        set_device_state(vrf_name, constants.LINK_UP)
    except agent_exc.NetworkInterfaceNotFound:
        create_interface(vrf_name, 'vrf', vrf_table=vrf_table,
                         state=constants.LINK_UP)


@ovn_bgp_agent.privileged.default.entrypoint
def ensure_bridge(bridge_name):
    try:
        set_device_state(bridge_name, constants.LINK_UP)
    except agent_exc.NetworkInterfaceNotFound:
        create_interface(bridge_name, 'bridge', br_stp_state=0,
                         state=constants.LINK_UP)


@ovn_bgp_agent.privileged.default.entrypoint
def ensure_vxlan(vxlan_name, vni, local_ip, dstport):
    try:
        set_device_state(vxlan_name, constants.LINK_UP)
    except agent_exc.NetworkInterfaceNotFound:
        # FIXME: Perhaps we need to set neigh_suppress on
        create_interface(vxlan_name, 'vxlan',
                         vxlan_id=vni,
                         vxlan_port=dstport,
                         vxlan_local=local_ip,
                         vxlan_learning=False,
                         state=constants.LINK_UP)


@ovn_bgp_agent.privileged.default.entrypoint
def ensure_veth(veth_name, veth_peer):
    try:
        set_device_state(veth_name, constants.LINK_UP)
    except agent_exc.NetworkInterfaceNotFound:
        create_interface(veth_name, 'veth', peer=veth_peer,
                         state=constants.LINK_UP)
    set_device_state(veth_peer, constants.LINK_UP)


@ovn_bgp_agent.privileged.default.entrypoint
def ensure_dummy_device(device):
    try:
        set_device_state(device, constants.LINK_UP)
    except agent_exc.NetworkInterfaceNotFound:
        create_interface(device, 'dummy', state=constants.LINK_UP)


@ovn_bgp_agent.privileged.default.entrypoint
def ensure_vlan_device_for_network(bridge, vlan_tag):
    vlan_device_name = '{}.{}'.format(
        bridge[:constants.OVN_VLAN_DEVICE_MAX_LENGTH],
        vlan_tag)
    try:
        set_device_state(vlan_device_name, constants.LINK_UP)
    except agent_exc.NetworkInterfaceNotFound:
        create_interface(vlan_device_name, 'vlan',
                         physical_interface=bridge,
                         vlan_id=vlan_tag,
                         state=constants.LINK_UP)


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
@ovn_bgp_agent.privileged.default.entrypoint
def set_master_for_device(device, master):
    try:
        dev_index = _get_link_id(ifname=device)
        master_index = _get_link_id(ifname=master)
        with pyroute2.IPRoute() as ipr:
            # Check if already associated to the master,
            # and associate it if not
            iface = ipr.link('get', index=dev_index)[0]
            if iface.get_attr('IFLA_MASTER') != master_index:
                ipr.link('set', index=dev_index, master=master_index)
    except IndexError:
        LOG.debug("No need to set %s on VRF %s, as one of them is deleted",
                  device, master)


@ovn_bgp_agent.privileged.default.entrypoint
def delete_device(device):
    try:
        delete_interface(device)
    except agent_exc.NetworkInterfaceNotFound:
        LOG.debug("Interfaces %s already deleted.", device)


@ovn_bgp_agent.privileged.default.entrypoint
def route_create(route):
    scope = route.pop('scope', 'link')
    route['scope'] = get_scope_name(scope)
    if 'family' not in route:
        route['family'] = constants.AF_INET
    _run_iproute_route('replace', **route)


@ovn_bgp_agent.privileged.default.entrypoint
def route_delete(route):
    scope = route.pop('scope', 'link')
    route['scope'] = get_scope_name(scope)
    if 'family' not in route:
        route['family'] = constants.AF_INET
    _run_iproute_route('del', **route)


@ovn_bgp_agent.privileged.default.entrypoint
def set_kernel_flag(flag, value):
    command = ["sysctl", "-w", "{}={}".format(flag, value)]
    try:
        return processutils.execute(*command)
    except Exception as e:
        LOG.error("Unable to execute %s. Exception: %s", command, e)
        raise


@ovn_bgp_agent.privileged.default.entrypoint
def delete_exposed_ips(ips, nic):
    for ip_address in ips:
        delete_ip_address(ip_address, nic)


@ovn_bgp_agent.privileged.default.entrypoint
def rule_create(rule):
    _run_iproute_rule('add', **rule)


@ovn_bgp_agent.privileged.default.entrypoint
def rule_delete(rule):
    _run_iproute_rule('del', **rule)


@ovn_bgp_agent.privileged.default.entrypoint
def delete_ip_rules(ip_rules):
    for rule_ip, rule_info in ip_rules.items():
        rule = l_net.create_rule_from_ip(rule_ip, int(rule_info['table']))
        _run_iproute_rule('del', **rule)


@ovn_bgp_agent.privileged.default.entrypoint
def add_ndp_proxy(ip, dev, vlan=None):
    # FIXME(ltomasbo): This should use pyroute instead but I didn't find
    # out how
    net_ip = str(ipaddress.IPv6Network(ip, strict=False).network_address)
    dev_name = dev
    if vlan:
        dev_name = f"{dev[:constants.OVN_VLAN_DEVICE_MAX_LENGTH]}.{vlan}"
    command = ["ip", "-6", "nei", "add", "proxy", net_ip, "dev", dev_name]
    try:
        return processutils.execute(*command)
    except Exception as e:
        LOG.error("Unable to execute %s. Exception: %s", command, e)
        raise


@ovn_bgp_agent.privileged.default.entrypoint
def del_ndp_proxy(ip, dev, vlan=None):
    # FIXME(ltomasbo): This should use pyroute instead but I didn't find
    # out how
    net_ip = str(ipaddress.IPv6Network(ip, strict=False).network_address)
    dev_name = dev
    if vlan:
        dev_name = f"{dev[:constants.OVN_VLAN_DEVICE_MAX_LENGTH]}.{vlan}"
    command = ["ip", "-6", "nei", "del", "proxy", net_ip, "dev", dev_name]
    env = dict(os.environ)
    env['LC_ALL'] = 'C'
    try:
        return processutils.execute(*command, env_variables=env)
    except Exception as e:
        if "No such file or directory" in e.stderr:
            # Already deleted
            return
        LOG.error("Unable to execute %s. Exception: %s", command, e)
        raise


@ovn_bgp_agent.privileged.default.entrypoint
def add_ip_to_dev(ip, nic):
    add_ip_address(ip, nic)


@ovn_bgp_agent.privileged.default.entrypoint
def del_ip_from_dev(ip, nic):
    delete_ip_address(ip, nic)


@ovn_bgp_agent.privileged.default.entrypoint
def add_ip_nei(ip, lladdr, dev):
    ip_version = l_net.get_ip_version(ip)
    family = common_utils.IP_VERSION_FAMILY_MAP[ip_version]
    _run_iproute_neigh('replace',
                       dev,
                       dst=ip,
                       lladdr=lladdr,
                       family=family,
                       state=ndmsg.states['permanent'])


@ovn_bgp_agent.privileged.default.entrypoint
def del_ip_nei(ip, lladdr, dev):
    ip_network = netaddr.IPNetwork(ip)
    family = common_utils.IP_VERSION_FAMILY_MAP[ip_network.version]

    _run_iproute_neigh('del',
                       dev,
                       dst=str(ip_network.ip),
                       lladdr=lladdr,
                       family=family,
                       state=ndmsg.states['permanent'])


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
def get_neigh_entries(device, ip_version, **kwargs):
    """Dump all neighbour entries.

    :param ip_version: IP version of entries to show (4 or 6)
    :param device: Device name to use in dumping entries
    :param kwargs: Callers add any filters they use as kwargs
    :return: a list of dictionaries, each representing a neighbour.
    The dictionary format is: {'dst': ip_address,
                               'lladdr': mac_address,
                               'device': device_name}
    """
    family = common_utils.IP_VERSION_FAMILY_MAP[ip_version]
    dump = _run_iproute_neigh('dump',
                              device,
                              family=family,
                              **kwargs)
    entries = []
    for entry in dump:
        attrs = dict(entry['attrs'])
        entries.append({'dst': attrs['NDA_DST'],
                        'lladdr': attrs.get('NDA_LLADDR'),
                        'device': device,
                        'state': NUD_STATES[entry['state']]})
    return entries


def add_unreachable_route(vrf_name):
    # NOTE(ltomasbo): This method is to set the default route for the table
    # (and hence default route for the VRF)
    # ip route add table 10 unreachable default metric 4278198272
    # Find vrf table.
    device = get_link_device(vrf_name)
    ifla_linkinfo = get_attr(device, 'IFLA_LINKINFO')
    ifla_data = get_attr(ifla_linkinfo, 'IFLA_INFO_DATA')
    vrf_table = get_attr(ifla_data, 'IFLA_VRF_TABLE')
    for ip_version in common_utils.IP_VERSION_FAMILY_MAP.values():
        kwargs = {'dst': 'default',
                  'family': ip_version,
                  'table': vrf_table,
                  'type': 'unreachable',
                  'scope': None,
                  'proto': 'boot',
                  'priority': 4278198272}
        route_create(kwargs)


@ovn_bgp_agent.privileged.default.entrypoint
def create_routing_table_for_bridge(table_number, bridge):
    with open('/etc/iproute2/rt_tables', 'a') as rt_tables:
        rt_tables.write('{} {}\n'.format(table_number, bridge))


def _translate_ip_device_exception(e, device):
    if e.code == errno.ENODEV:
        raise agent_exc.NetworkInterfaceNotFound(device=device)
    if e.code == errno.EOPNOTSUPP:
        raise agent_exc.InterfaceOperationNotSupported(device=device)
    if e.code == errno.EINVAL:
        raise agent_exc.InvalidArgument(device=device)
    if e.code == errno.EEXIST:
        raise agent_exc.InterfaceAlreadyExists(device=device)
    raise e


def _translate_ip_addr_exception(e, ip, device):
    if e.code == errno.EEXIST:
        raise agent_exc.IpAddressAlreadyExists(ip=ip, device=device)
    if e.code == errno.EADDRNOTAVAIL:
        LOG.debug('No need to delete IP address %s on dev %s as it does '
                  'not exist', ip, device)
        return
    raise e


def _translate_ip_route_exception(e, kwargs):
    if e.code == errno.EEXIST:  # Already exists
        LOG.debug("Route %s already exists.", kwargs)
        return
    if e.code == errno.ENOENT or e.code == errno.ESRCH:  # Not found
        LOG.debug("Route already deleted: %s", kwargs)
        return
    raise e


def _translate_ip_rule_exception(e, kwargs):
    if e.code == errno.EEXIST:  # Already exists
        LOG.debug("Rule %s already exists.", kwargs)
        return
    if e.code == errno.ENOENT:  # Not found
        LOG.debug("Rule already deleted: %s", kwargs)
        return
    raise e


def get_attr(pyroute2_obj, attr_name):
    """Get an attribute in a pyroute object

    pyroute2 object attributes are stored under a key called 'attrs'. This key
    contains a tuple of tuples. E.g.:
      pyroute2_obj = {'attrs': (('TCA_KIND': 'htb'),
                                ('TCA_OPTIONS': {...}))}

    :param pyroute2_obj: (dict) pyroute2 object
    :param attr_name: (string) first value of the tuple we are looking for
    :return: (object) second value of the tuple, None if the tuple doesn't
             exist
    """
    rule_attrs = pyroute2_obj.get('attrs', [])
    for attr in (attr for attr in rule_attrs if attr[0] == attr_name):
        return attr[1]


def make_serializable(value):
    """Make a pyroute2 object serializable

    This function converts 'netlink.nla_slot' object (key, value) in a list
    of two elements.
    """
    def _ensure_string(value):
        return value.decode() if isinstance(value, bytes) else value

    if isinstance(value, list):
        return [make_serializable(item) for item in value]
    elif isinstance(value, pyroute_netlink.nla_slot):
        return [_ensure_string(value[0]), make_serializable(value[1])]
    elif isinstance(value, pyroute_netlink.nla_base):
        return make_serializable(value.dump())
    elif isinstance(value, dict):
        return {_ensure_string(key): make_serializable(data)
                for key, data in value.items()}
    elif isinstance(value, tuple):
        return tuple(make_serializable(item) for item in value)
    return _ensure_string(value)


def _get_link_id(ifname, raise_exception=True):
    ifname = ifname[:n_const.DEVICE_NAME_MAX_LEN]
    with iproute.IPRoute() as ip:
        link_id = ip.link_lookup(ifname=ifname)
    if not link_id or len(link_id) < 1:
        if raise_exception:
            raise agent_exc.NetworkInterfaceNotFound(device=ifname)
        LOG.debug('Interface %(dev)s not found', {'dev': ifname})
        return
    return link_id[0]


@ovn_bgp_agent.privileged.default.entrypoint
def get_link_id(device):
    return _get_link_id(device, raise_exception=False)


def get_link_state(device_name):
    device = get_link_device(device_name)
    return device['state'] if device else None


def get_link_device(device_name):
    device_name = device_name[:n_const.DEVICE_NAME_MAX_LEN]
    for device in get_link_devices():
        if get_attr(device, 'IFLA_IFNAME') == device_name:
            return device


@ovn_bgp_agent.privileged.default.entrypoint
def get_bridge_vlans(device_name):
    index = _get_link_id(device_name, raise_exception=False)
    if not index:
        LOG.debug("OVS Bridge %s deleted, no need to get information about "
                  "associated vlan devices", device_name)

    vlan_devices = get_link_devices(link=index)
    vlans = []
    for vlan_device in vlan_devices:
        ifla_linkinfo = get_attr(vlan_device, 'IFLA_LINKINFO')
        if ifla_linkinfo:
            ifla_data = get_attr(ifla_linkinfo, 'IFLA_INFO_DATA')
            if ifla_data:
                vlans.append(get_attr(ifla_data, 'IFLA_VLAN_ID'))
    return vlans


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
@ovn_bgp_agent.privileged.default.entrypoint
def get_link_devices(**kwargs):
    """List interfaces in a namespace

    :return: (list) interfaces in a namespace
    """
    index = kwargs.pop('index') if 'index' in kwargs else 'all'
    try:
        with iproute.IPRoute() as ip:
            return make_serializable(ip.get_links(index, **kwargs))
    except OSError:
        raise


def _run_iproute_link(command, ifname, **kwargs):
    try:
        with iproute.IPRoute() as ip:
            idx = _get_link_id(ifname)
            return ip.link(command, index=idx, **kwargs)
    except netlink_exceptions.NetlinkError as e:
        _translate_ip_device_exception(e, ifname)


def _run_iproute_addr(command, device, **kwargs):
    try:
        with iproute.IPRoute() as ip:
            idx = _get_link_id(device)
            return ip.addr(command, index=idx, **kwargs)
    except netlink_exceptions.NetlinkError as e:
        _translate_ip_addr_exception(e, ip=kwargs['address'], device=device)


def _run_iproute_route(command, **kwargs):
    try:
        with iproute.IPRoute() as ip:
            ip.route(command, **kwargs)
    except netlink_exceptions.NetlinkError as e:
        _translate_ip_route_exception(e, kwargs)


def _run_iproute_rule(command, **kwargs):
    try:
        with iproute.IPRoute() as ip:
            ip.rule(command, **kwargs)
    except netlink_exceptions.NetlinkError as e:
        _translate_ip_rule_exception(e, kwargs)


def _run_iproute_neigh(command, device, **kwargs):
    try:
        with iproute.IPRoute() as ip:
            idx = _get_link_id(device)
            return ip.neigh(command, ifindex=idx, **kwargs)
    except agent_exc.NetworkInterfaceNotFound:
        LOG.debug("No need to %s nei for dev %s as it does not exists",
                  command, device)


def _run_iproute_brport(command, ifname, **kwargs):
    try:
        with iproute.IPRoute() as ip:
            idx = _get_link_id(ifname)
            return ip.brport(command, index=idx, **kwargs)
    except netlink_exceptions.NetlinkError as e:
        _translate_ip_device_exception(e, ifname)


@ovn_bgp_agent.privileged.default.entrypoint
def create_interface(ifname, kind, **kwargs):
    ifname = ifname[:n_const.DEVICE_NAME_MAX_LEN]
    try:
        with iproute.IPRoute() as ip:
            physical_interface = kwargs.pop('physical_interface', None)
            if physical_interface:
                link_key = 'vxlan_link' if kind == 'vxlan' else 'link'
                kwargs[link_key] = _get_link_id(physical_interface)
            ip.link("add", ifname=ifname, kind=kind, **kwargs)
    except netlink_exceptions.NetlinkError as e:
        _translate_ip_device_exception(e, ifname)


@ovn_bgp_agent.privileged.default.entrypoint
def delete_interface(ifname, **kwargs):
    ifname = ifname[:n_const.DEVICE_NAME_MAX_LEN]
    _run_iproute_link('del', ifname, **kwargs)


@ovn_bgp_agent.privileged.default.entrypoint
def set_link_attribute(ifname, **kwargs):
    ifname = ifname[:n_const.DEVICE_NAME_MAX_LEN]
    _run_iproute_link("set", ifname, **kwargs)


@ovn_bgp_agent.privileged.default.entrypoint
def set_brport_attribute(ifname, **kwargs):
    _run_iproute_brport("set", ifname, **kwargs)


@ovn_bgp_agent.privileged.default.entrypoint
def add_ip_address(ip_address, ifname, prefixlen=None, **kwargs):
    ifname = ifname[:n_const.DEVICE_NAME_MAX_LEN]
    net = netaddr.IPNetwork(ip_address)
    ip_version = l_net.get_ip_version(ip_address)
    address = str(net.ip)
    if not prefixlen:
        prefixlen = 32 if ip_version == 4 else 128
    family = common_utils.IP_VERSION_FAMILY_MAP[ip_version]
    _run_iproute_addr('add',
                      ifname,
                      address=address,
                      mask=prefixlen,
                      family=family)


@ovn_bgp_agent.privileged.default.entrypoint
def delete_ip_address(ip_address, ifname, prefixlen=None, **kwargs):
    ifname = ifname[:n_const.DEVICE_NAME_MAX_LEN]
    net = netaddr.IPNetwork(ip_address)
    ip_version = l_net.get_ip_version(ip_address)
    address = str(net.ip)
    if not prefixlen:
        prefixlen = 32 if ip_version == 4 else 128
    family = common_utils.IP_VERSION_FAMILY_MAP[ip_version]
    _run_iproute_addr("delete",
                      ifname,
                      address=address,
                      mask=prefixlen,
                      family=family,
                      **kwargs)


@ovn_bgp_agent.privileged.default.entrypoint
def get_ip_addresses(**kwargs):
    """List of IP addresses in a namespace

    :return: (tuple) IP addresses in a namespace
    """
    with iproute.IPRoute() as ip:
        return make_serializable(ip.get_addr(**kwargs))


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
@ovn_bgp_agent.privileged.default.entrypoint
def list_ip_routes(ip_version, device=None, table=None, **kwargs):
    """List IP routes"""
    kwargs['family'] = common_utils.IP_VERSION_FAMILY_MAP[ip_version]
    if device:
        kwargs['oif'] = _get_link_id(device)
    if table:
        kwargs['table'] = int(table)
    with iproute.IPRoute() as ip:
        return make_serializable(ip.route('show', **kwargs))


@ovn_bgp_agent.privileged.default.entrypoint
def list_ip_rules(ip_version, **kwargs):
    """List all IP rules"""
    with iproute.IPRoute() as ip:
        return make_serializable(ip.get_rules(
            family=common_utils.IP_VERSION_FAMILY_MAP[ip_version], **kwargs))



# Copyright 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast

from oslo_config import cfg
from oslo_log import log as logging

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import driver_utils
from ovn_bgp_agent.drivers.openstack.utils import evpn
from ovn_bgp_agent.drivers.openstack.utils import ovn
from ovn_bgp_agent.drivers.openstack.utils import ovs
from ovn_bgp_agent import exceptions as agent_exc
from ovn_bgp_agent.utils import helpers
from ovn_bgp_agent.utils import linux_net


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def ensure_base_wiring_config(idl, ovs_idl, ovn_idl=None, routing_tables={}):
    if CONF.exposing_method == constants.EXPOSE_METHOD_UNDERLAY:
        return _ensure_base_wiring_config_underlay(idl, ovs_idl,
                                                   routing_tables)
    elif CONF.exposing_method == constants.EXPOSE_METHOD_VRF:  # Type 5 evpn
        return _ensure_base_wiring_config_evpn(idl, ovs_idl)
    elif CONF.exposing_method == constants.EXPOSE_METHOD_OVN:
        return _ensure_base_wiring_config_ovn(ovs_idl, ovn_idl)

    raise agent_exc.UnsupportedWiringConfig(method=CONF.exposing_method)


def _ensure_base_wiring_config_underlay(idl, ovs_idl, routing_tables):
    # Get bridge mappings: xxxx:br-ex,yyyy:br-ex2
    bridge_mappings = ovs_idl.get_ovn_bridge_mappings()

    ovn_bridge_mappings = {}
    flows_info = {}
    for bridge_index, bridge_mapping in enumerate(bridge_mappings, 1):
        network, bridge = helpers.parse_bridge_mapping(bridge_mapping)
        if not network:
            continue
        ovn_bridge_mappings[network] = bridge

        linux_net.ensure_routing_table_for_bridge(
            routing_tables, bridge, CONF.bgp_vrf_table_id)
        vlan_tags = idl.get_network_vlan_tag_by_network_name(network)

        for vlan_tag in vlan_tags:
            linux_net.ensure_vlan_device_for_network(bridge,
                                                     vlan_tag)

        linux_net.ensure_arp_ndp_enabled_for_bridge(bridge,
                                                    bridge_index,
                                                    vlan_tags)
        if not flows_info.get(bridge):
            mac = linux_net.get_interface_address(bridge)
            flows_info[bridge] = {'mac': mac, 'in_port': set()}
            flows_info[bridge]['in_port'] = ovs.get_ovs_patch_ports_info(
                bridge)
            ovs.ensure_mac_tweak_flows(bridge, mac,
                                       flows_info[bridge]['in_port'],
                                       constants.OVS_RULE_COOKIE)
    return ovn_bridge_mappings, flows_info


def _ensure_base_wiring_config_evpn(idl: 'ovn.OvsdbNbOvnIdl|ovn.OvsdbSbOvnIdl',
                                    ovs_idl: 'ovs.OvsIdl',
                                    mode=constants.OVN_EVPN_TYPE_L3):
    # This method will create the bridge mappings and make sure the
    # vrf bridge and vlans are provisioned

    # Get bridge mappings: xxxx:br-ex,yyyy:br-ex2
    bridge_mappings = ovs_idl.get_ovn_bridge_mappings()

    ovn_bridge_mappings = {}
    flows_info = {}  # dictionary to use for mappings with vrf's
    for bridge_mapping in bridge_mappings:
        try:
            # for example: physnet1, br-ex
            network, bridge = bridge_mapping.split(":", 1)
        except ValueError:
            LOG.debug('Invalid bridge mapping: %s', bridge_mapping)
            continue

        ovn_bridge_mappings[network] = bridge
        LOG.debug('Setup EVPN base wiring for network %s on bridge %s',
                  network, bridge)

        # Make sure the bridge exists
        ovs_idl.idl_ovs.add_br(bridge).execute(check_error=True)

        if bridge not in flows_info:
            flows_info[bridge] = {
                'mac': linux_net.get_interface_address(bridge),
                'in_port': ovs.get_ovs_patch_ports_info(bridge),
                'evpn': {}
            }

        # Find all provider networks, and create the vrf's
        localnet_ports = list(idl.get_localnet_ports_by_network_name(network))
        if not localnet_ports:
            LOG.debug('No localnet ports found for network %s', network)
            continue

        provnets = idl.get_bgpvpn_networks_for_ports(localnet_ports,
                                                     vpn_type=mode)
        if not provnets:
            LOG.debug('No provider networks found for %s %s',
                      constants.OVN_EVPN_TYPE_EXT_ID_KEY, mode)
            continue

        for ls in provnets:
            LOG.info('Network %s (settings: %s)', ls.name, ls.external_ids)

            if constants.OVN_EVPN_VNI_EXT_ID_KEY not in ls.external_ids:
                LOG.warning('Skipped, VNI required for EVPN VRF setup')
                continue

            evpn_opts = {}
            for opt, ext_id_key in constants.EVPN_EXT_ID_MAPPING.items():
                evpn_opts[opt] = ast.literal_eval(
                    ls.external_ids.get(ext_id_key, '[]')
                )

            # Create or return the EVPN bridge
            evpn_bridge = evpn.setup(
                ovs_bridge=bridge,
                vni=ls.external_ids[constants.OVN_EVPN_VNI_EXT_ID_KEY],
                evpn_opts=evpn_opts,
                mode=mode,
                ovs_flows=flows_info,
            )

            # Connect all VLAN interfaces to this VRF and gather dhcp
            # options to be configured for l3 mode.
            evpn_dev, dhcp_opts = _ensure_evpn_vlan_dev(ls, localnet_ports,
                                                        evpn_bridge,
                                                        flows_info, bridge)

            if dhcp_opts and evpn_dev:
                evpn_dev.process_dhcp_opts(dhcp_opts)

    return ovn_bridge_mappings, flows_info


def _ensure_evpn_vlan_dev(ls, localnet_ports, evpn_bridge, flows_info, bridge):
    evpn_dev = None
    dhcp_opts = set()
    for port in ls.ports:
        if port not in localnet_ports:
            if port.dhcpv4_options:
                dhcp_opts.update(port.dhcpv4_options)
            if port.dhcpv6_options:
                dhcp_opts.update(port.dhcpv6_options)
            continue

        LOG.info('VLAN tag %s', driver_utils.get_port_vlan(port))
        evpn_dev = evpn_bridge.connect_vlan(port)

        flows_info[bridge]['evpn'][str(evpn_dev.vlan_tag)] = evpn_bridge

    return evpn_dev, dhcp_opts


def _ensure_base_wiring_config_ovn(ovs_idl, ovn_idl):
    """Base configuration for extra OVN cluster instead of kernel networking

    This function is in charge of the steps to ensure the in-node OVN cluster
    is properly configured by:
    1. Get the information about the OpenStack provider bridge(s) and the flows
       info, such as the mac and the in_port
    2. Add the egress ovs flows so that the destination mac is the one on the
       extra ovn-cluster, in the LR
    3. Create the LR in the in-node OVN cluster
    4. Create the LR policy in the in-node OVN cluster to redirect the traffic
       (with ECMP support) to the nexthops
    5. Create the LR routes in the in-node OVN cluster to route any traffic the
       peers IPs
    6. Create the LS (+ Localnet port) for the connection between the router
       and the OpenStack OVN networks. then it connects it to the LR
    7. Create the LS (+ Localnet port) for the connection between the router
       and the external network. Then it connects it to the LR
    8. Create the ingress_flow at the external OVN provider bridges to
       redirect the needed traffic to the in-cluster OVN networks

    :param ovs_idl: The idl to communicate with local ovs DB
    :param ovn_idl: The idl to communicate with local (in-node) NB DB
    :return: ovn_bridge_mappings (network and bridges association) and the
             flows_info per bridge
    """
    # OpenStack Egress part
    # Get bridge mappings: xxxx:br-ex,yyyy:br-ex2
    bridge_mappings = ovs_idl.get_ovn_bridge_mappings()
    ovn_bridge_mappings = {}
    flows_info = {}
    for bridge_mapping in bridge_mappings:
        network, bridge = helpers.parse_bridge_mapping(bridge_mapping)
        if not network:
            continue
        ovn_bridge_mappings[network] = bridge

        if not flows_info.get(bridge):
            mac = linux_net.get_interface_address(bridge)
            flows_info[bridge] = {'mac': mac, 'in_port': set()}
            flows_info[bridge]['in_port'] = ovs.get_ovs_patch_ports_info(
                bridge)
        _ensure_egress_flows(bridge, flows_info[bridge]['in_port'])

    # Extra OVN cluster configuration
    provider_cidrs = CONF.local_ovn_cluster.provider_networks_pool_prefixes

    # LR
    cmds = []
    cmds.extend(_ensure_ovn_router(ovn_idl))
    # FIXME(ltomasbo): we need to firsts create the router and then the
    # policies and routes in a different transaction until ovsdbapp
    # allows it to do it in one transaction. Once that happen the next
    # 2 lines can be removed
    _execute_commands(ovn_idl, cmds)
    cmds = []
    cmds.extend(_ensure_ovn_policies(ovn_idl, CONF.local_ovn_cluster.peer_ips))

    ovs_idl.get_ovn_bridge_mappings(
        bridge=CONF.local_ovn_cluster.bgp_chassis_id)

    bfds = _ensure_bfds(ovn_idl, zip(
        CONF.local_ovn_cluster.peer_ips,
        ovs_idl.get_ovn_bridge_mappings(
            bridge=CONF.local_ovn_cluster.bgp_chassis_id)))
    cmds.extend(
        _ensure_ovn_routes(ovn_idl, CONF.local_ovn_cluster.peer_ips, bfds))
    # Creation of all router related cmds in a single transaction
    _execute_commands(ovn_idl, cmds)

    # LS
    bgp_bridge_mappings = ovs_idl.get_ovn_bridge_mappings(
        bridge=CONF.local_ovn_cluster.bgp_chassis_id)
    for bridge_mapping in bgp_bridge_mappings:
        network, bridge = helpers.parse_bridge_mapping(bridge_mapping)
        if not network:
            continue
        # Create LS + Localnet port on it
        _ensure_ovn_switch(ovn_idl, network)

        # differentiate between internal LS (connecting to OpenStack)
        # and external LS (connecting to the NICs)
        if bridge in ovn_bridge_mappings.values():
            # Internal Bridge connecting to OpenStack OVN cluster
            _ensure_ovn_network_link_internal(ovn_idl, network, provider_cidrs)
        else:
            ip, mac = linux_net.get_nic_info(bridge)
            # External Bridge connecting to the external networks
            _ensure_ovn_network_link_external(ovn_idl, network, ip, mac)
            _ensure_ingress_flows(bridge, mac, network, provider_cidrs)
            _ensure_fdb_entry(bridge, mac)

    return ovn_bridge_mappings, flows_info


def _ensure_bfds(ovn_idl, nic_infos):
    bfds = []
    for peer_ip, bm in nic_infos:
        network, bridge = helpers.parse_bridge_mapping(bm)
        # the port name is hardcoded
        r_port_name = "{}-{}".format(
            constants.OVN_CLUSTER_ROUTER, network)
        bfds.append(
            ovn_idl.bfd_add(
                logical_port=r_port_name, dst_ip=peer_ip,
                may_exist=True).execute(check_error=True))

    return bfds


def _ensure_ovn_router(ovn_idl):
    return [ovn_idl.lr_add(constants.OVN_CLUSTER_ROUTER, may_exist=True)]


def _ensure_ovn_switch(ovn_idl, switch_name):
    ovn_idl.ls_add(switch_name, may_exist=True).execute(check_error=True)

    # Add localnet port to them
    localnet_port = "{}-localnet".format(switch_name)
    options = {'network_name': switch_name}
    cmds = _ensure_lsp_cmds(ovn_idl, localnet_port, switch_name, 'localnet',
                            'unknown', **options)
    _execute_commands(ovn_idl, cmds)


def _execute_commands(idl, cmds):
    with idl.transaction(check_error=True) as txn:
        for command in cmds:
            txn.add(command)


def _ensure_ovn_network_link_internal(ovn_idl, switch_name, provider_cidrs):
    """Configure connecting LR and LSs internal side

    1. Creates LRP to connect to the internal switch
    2. If networks (provider_cidrs) are different, adding the new networks
    3. Create LSP related to the LRP with the right options, including the
       arp_proxy
    4. Bind the LRP to the local chassis

    :param ovn_idl: The idl to communicate with local (in-node) NB DB
    :param switch_name: the name of the logical switch to configure
    :param provider_cidrs: CIDRs to configure the networks of the
                           LRP, as well as to configure the ARP
                           proxy on the internal LSP
    """
    cmds = []
    # Connect BGP router to the internal logical switch
    r_port_name = "{}-openstack".format(constants.OVN_CLUSTER_ROUTER)
    try:
        ovn_idl.lrp_add(constants.OVN_CLUSTER_ROUTER, r_port_name,
                        constants.OVN_CLUSTER_ROUTER_INTERNAL_MAC,
                        provider_cidrs, peer=[], may_exist=True).execute(
            check_error=True)
    except RuntimeError as rte:
        # TODO(ltomasbo): Change OVSDBAPP to return a different error for
        # this to avoid having to compare strings as this is error prone
        networks_message = 'with different networks'
        if networks_message not in str(rte):
            raise
        # Trying to sync the networks by adding them
        cmds.append(ovn_idl.lrp_add_networks(r_port_name, provider_cidrs,
                                             may_exist=True))

    s_port_name = "openstack-{}".format(constants.OVN_CLUSTER_ROUTER)
    # NOTE(ltomasbo): require v23.06.0 so that proxy-arp works as expected.
    # If older version the provider_cidrs should contain all the provider
    # network cidrs, pointing to the gateway IP of the network.
    cidrs = ','.join(provider_cidrs) if provider_cidrs else '0.0.0.0/0'
    options = {'router-port': r_port_name, 'arp_proxy': cidrs}
    cmds.extend(_ensure_lsp_cmds(ovn_idl, s_port_name, switch_name,
                                 'router', 'router', **options))
    # bind to local chassis
    # ovn-nbctl lrp-set-gateway-chassis  bgp-router-public bgp 1
    cmds.append(ovn_idl.lrp_set_gateway_chassis(
        r_port_name, CONF.local_ovn_cluster.bgp_chassis_id, 1))

    _execute_commands(ovn_idl, cmds)


def _ensure_ovn_network_link_external(ovn_idl, switch_name, ip, mac):
    """Configure connecting LR and LSs external side

    1. Creates LRP to connect to the external switch
    2. If networks (ip) is different than the nic network add the nic network
       and remove the extra ones
    3. Create LSP related to the LRP with the right options

    :param ovn_idl: The idl to communicate with local (in-node) NB DB
    :param switch_name: the name of the logical switch to configure
    :param ip: IP to configure in the LRP connected to the external switch
    :param mac: MAC to configure in the LRP connected to the external switch
    """
    cmds = []
    r_port_name = "{}-{}".format(constants.OVN_CLUSTER_ROUTER, switch_name)

    # LRP
    try:
        ovn_idl.lrp_add(constants.OVN_CLUSTER_ROUTER, r_port_name,
                        mac, [ip], peer=[], may_exist=True).execute(
            check_error=True)
    except RuntimeError as rte:
        # TODO(ltomasbo): Change OVSDBAPP to return a different error for
        # this to avoid having to compare strings as this is error prone
        networks_message = 'with different networks'
        if networks_message not in str(rte):
            raise
        # Trying to sync the networks by adding them
        cmds.append(ovn_idl.lrp_add_networks(r_port_name,
                                             [ip],
                                             may_exist=True))
        lrp = ovn_idl.lrp_get(r_port_name).execute(check_error=True)
        for net in lrp.networks:
            if net != ip:
                cmds.append(ovn_idl.lrp_del_networks(r_port_name,
                                                     [net],
                                                     if_exists=True))
    # LSP
    s_port_name = "{}-{}".format(switch_name, constants.OVN_CLUSTER_ROUTER)
    options = {'router-port': r_port_name}
    cmds.extend(_ensure_lsp_cmds(ovn_idl, s_port_name, switch_name,
                                 'router', 'router', **options))

    # bind to local chassis
    # ovn-nbctl lrp-set-gateway-chassis  bgp-router-public bgp 1
    cmds.append(ovn_idl.lrp_set_gateway_chassis(
        r_port_name, CONF.local_ovn_cluster.bgp_chassis_id, 1))

    _execute_commands(ovn_idl, cmds)


def _ensure_lsp_cmds(ovn_idl, port_name, switch, port_type, addresses,
                     **options):
    cmds = []
    cmds.append(ovn_idl.lsp_add(switch, port_name, may_exist=True))
    cmds.append(ovn_idl.lsp_set_type(port_name, port_type))
    cmds.append(ovn_idl.lsp_set_addresses(port_name,
                                          addresses=[addresses]))
    cmds.append(ovn_idl.lsp_set_options(port_name, **options))
    return cmds


def _ensure_ovn_policies(ovn_idl, next_hops):
    priority = 10
    match = 'inport=="{}-openstack"'.format(constants.OVN_CLUSTER_ROUTER)
    action = 'reroute'
    columns = {}
    if len(next_hops) > 1:
        columns = {'nexthops': next_hops}
    elif len(next_hops) == 1:
        columns = {'nexthop': next_hops[0]}

    return [ovn_idl.lr_policy_add(constants.OVN_CLUSTER_ROUTER, priority,
                                  match, action, may_exist=True, **columns)]


def _ensure_ovn_routes(ovn_idl, peer_ips, bfds):
    prefix = '0.0.0.0/0'
    cmds = []
    for bfd, ip in zip(bfds, peer_ips):
        cmds.append(ovn_idl.lr_route_add(constants.OVN_CLUSTER_ROUTER, prefix,
                                         ip, ecmp=True, may_exist=True,
                                         bfd=bfd.uuid))
    return cmds


def _ensure_ingress_flows(bridge, mac, switch_name, provider_cidrs):
    # incomming traffic flows
    # patch=`ovs-ofctl show br-ex | grep patch | cut -d "("  -f1 | xargs`
    # ovs-ofctl add-flow br-ex
    #    "cookie=0xbadcaf2,ip,nw_dst=$PROVIDER_NET,in_port=enp2s0,priority=100,
    #     actions=mod_dl_dst:$ENP2S0_MAC,output=$patch"
    if not provider_cidrs:
        return
    patch_port_prefix = 'patch-{}-'.format(switch_name)
    patch_ports = ovs.get_ovs_patch_ports_info(bridge,
                                               prefix=patch_port_prefix)
    if not patch_ports:
        return
    bridge_ports = set(ovs.get_ovs_ports_info(bridge))
    external_nic = list(bridge_ports.intersection(
        set(CONF.local_ovn_cluster.external_nics)))

    if not external_nic:
        LOG.warning("NIC ports (%s) not found for bridge %s. Not possible to "
                    "create the ingress flows. It will be retried if "
                    "reconcile cycle is not disabled",
                    CONF.local_ovn_cluster.external_nics, bridge)
        return
    else:
        # only one external_nic expected per bridge
        external_nic = external_nic[0]

    for provider_cidr in provider_cidrs:
        ip_version = linux_net.get_ip_version(provider_cidr)
        if ip_version == constants.IP_VERSION_6:
            ip = 'ipv6'
        else:
            ip = 'ip'
        flow = (
            "cookie={},priority=1000,{},nw_dst={},in_port={},"
            " actions=mod_dl_dst:{},output={}".format(
                constants.OVS_RULE_COOKIE, ip, provider_cidr, external_nic,
                mac, patch_ports[0]))
        ovs.ensure_flow(bridge, flow)


def _ensure_fdb_entry(bridge, mac):
    """Ensure static FDB entries for internal ports of a fabric bridge.

    Fabric-facing bridges need to have a static MAC entry to avoid dynamic
    re-learning in their FDB as this will lead to drop flows being generated
    by OVS due to flow table misses.
    """
    # The bridge name is also the name of its internal port. The MAC address
    # is the MAC address of the bridge itself.
    ovs.add_fdb_entry(bridge, bridge, mac, 0)


def _ensure_egress_flows(bridge, patch_ports):
    # outcomming traffic flows
    # patch=`ovs-ofctl show br-provider | grep patch | grep provnet |
    #     cut -d "("  -f1 | xargs`
    # ovs-ofctl add-flow br-provider "cookie=0xbadcaf3,ip,in_port=$patch,
    #     actions=mod_dl_dst:$ROUTER_MAC,NORMAL"
    for patch_port in patch_ports:
        flow = (
            "cookie={},priority=1000,ip,in_port={},"
            " actions=mod_dl_dst:{},NORMAL".format(
                constants.OVS_RULE_COOKIE, patch_port,
                constants.OVN_CLUSTER_ROUTER_INTERNAL_MAC))
        flow_v6 = (
            "cookie={},priority=1000,ipv6,in_port={},"
            " actions=mod_dl_dst:{},NORMAL".format(
                constants.OVS_RULE_COOKIE, patch_port,
                constants.OVN_CLUSTER_ROUTER_INTERNAL_MAC))
        ovs.ensure_flow(bridge, flow)
        ovs.ensure_flow(bridge, flow_v6)


def cleanup_wiring(idl, bridge_mappings, ovs_flows, exposed_ips,
                   routing_tables, routing_tables_routes):
    if CONF.exposing_method == constants.EXPOSE_METHOD_UNDERLAY:
        return _cleanup_wiring_underlay(idl, bridge_mappings, ovs_flows,
                                        exposed_ips, routing_tables,
                                        routing_tables_routes)
    elif CONF.exposing_method == constants.EXPOSE_METHOD_VRF:
        return _cleanup_wiring_evpn(ovs_flows, routing_tables_routes)
    elif CONF.exposing_method == constants.EXPOSE_METHOD_OVN:
        return _cleanup_wiring_ovn(exposed_ips)


def _cleanup_extra_ips(exposed_ips: list[str]) -> set[str]:
    current_ips = set(linux_net.get_exposed_ips(CONF.bgp_nic))
    expected_ips = {ip for ip_dict in exposed_ips.values()
                    for ip in ip_dict.keys()}
    ips_to_delete = current_ips - expected_ips
    if ips_to_delete:
        LOG.warning(f"Removing IPs {ips_to_delete} from {CONF.bgp_nic} as "
                    "they are not expected to be exposed.")
        linux_net.delete_exposed_ips(ips_to_delete, CONF.bgp_nic)
    return expected_ips


def _cleanup_wiring_ovn(exposed_ips):
    """Cleanup IPs that are not expected to be exposed on the bgp_nic.

    This may be needed in case ovn-bgp-agent misses updates from the OVN NB
    database due to temporary downtime, networking issues between the
    ovn-bgp-agent and the NB DB or otherwise.
    """
    _cleanup_extra_ips(exposed_ips)
    # TODO(dmitriis): clean up other state: in the local OVN NB: logical
    # router policies, logical router routes, ARP proxies in logical switch
    # ports, static mac bindings.


def _cleanup_wiring_underlay(idl, bridge_mappings, ovs_flows, exposed_ips,
                             routing_tables, routing_tables_routes):
    expected_ips = _cleanup_extra_ips(exposed_ips)
    extra_routes = {}
    for bridge in bridge_mappings.values():
        extra_routes[bridge] = (
            linux_net.get_extra_routing_table_for_bridge(routing_tables,
                                                         bridge))
        # delete extra ovs flows
        ovs.remove_extra_ovs_flows(ovs_flows, bridge,
                                   constants.OVS_RULE_COOKIE)

    # get rules and delete the old ones
    ovn_ip_rules = linux_net.get_ovn_ip_rules(routing_tables.values())
    if ovn_ip_rules:
        for ip in expected_ips:
            if len(ip.split("/")) == 1:
                ip_version = linux_net.get_ip_version(ip)
                if ip_version == constants.IP_VERSION_6:
                    ip_dst = "{}/128".format(ip)
                else:
                    ip_dst = "{}/32".format(ip)
            else:
                ip_dst = ip
            ovn_ip_rules.pop(ip_dst, None)
    linux_net.delete_ip_rules(ovn_ip_rules)

    # remove all the extra routes not needed
    linux_net.delete_bridge_ip_routes(routing_tables, routing_tables_routes,
                                      extra_routes)

    # delete leaked vlan devices from previous vlan provider networks
    delete_vlan_devices_leftovers(idl, bridge_mappings)


def delete_vlan_devices_leftovers(idl, bridge_mappings):
    vlan_tags = idl.get_network_vlan_tags()
    ovs_devices = set(bridge_mappings.values())
    for ovs_device in ovs_devices:
        vlans = linux_net.get_bridge_vlans(ovs_device)
        for vlan in vlans:
            if vlan and vlan not in vlan_tags:
                linux_net.delete_vlan_device_for_network(ovs_device, vlan)


def _cleanup_wiring_evpn(ovs_flows, routing_tables_routes):
    for flow_conf in ovs_flows.values():
        for vlan, evpn_bridge in flow_conf.get('evpn', {}).items():
            LOG.debug('Running cleanup for vrf %s vlan %s',
                      evpn_bridge.vrf_name, vlan)
            evpn_bridge.get_vlan(vlan).cleanup_excessive_routes(
                routing_tables_routes
            )
    return True


def wire_provider_port(routing_tables_routes, ovs_flows, port_ips,
                       bridge_device, bridge_vlan, localnet, routing_table,
                       proxy_cidrs, mac=None, ovn_idl=None):
    if CONF.exposing_method == constants.EXPOSE_METHOD_UNDERLAY:
        return _wire_provider_port_underlay(routing_tables_routes, ovs_flows,
                                            port_ips, bridge_device,
                                            bridge_vlan, localnet,
                                            routing_table, proxy_cidrs,
                                            lladdr=mac)
    elif CONF.exposing_method == constants.EXPOSE_METHOD_VRF:
        return _wire_provider_port_evpn(routing_tables_routes, ovs_flows,
                                        port_ips, bridge_device,
                                        bridge_vlan, localnet,
                                        proxy_cidrs,
                                        mac=mac)
    elif CONF.exposing_method == constants.EXPOSE_METHOD_OVN:
        # We need to add a static mac binding due to proxy-arp issue in
        # core ovn that would reply on the incomming traffic from the LR,
        # while it should not
        return _wire_provider_port_ovn(ovn_idl, port_ips, mac)


def unwire_provider_port(routing_tables_routes, port_ips, bridge_device,
                         bridge_vlan, routing_table, proxy_cidrs, mac=None,
                         ovn_idl=None):
    if CONF.exposing_method == constants.EXPOSE_METHOD_UNDERLAY:
        return _unwire_provider_port_underlay(routing_tables_routes, port_ips,
                                              bridge_device, bridge_vlan,
                                              routing_table, proxy_cidrs,
                                              lladdr=mac)
    elif CONF.exposing_method == constants.EXPOSE_METHOD_VRF:
        return _unwire_provider_port_evpn(routing_tables_routes, port_ips,
                                          bridge_device, bridge_vlan,
                                          mac)
    elif CONF.exposing_method == constants.EXPOSE_METHOD_OVN:
        # We need to remove thestatic mac binding added due to proxy-arp issue
        # in core ovn that would reply on the incomming traffic from the LR,
        # while it should not
        return _unwire_provider_port_ovn(ovn_idl, port_ips)


def _ensure_updated_mac_tweak_flows(localnet, bridge_device, ovs_flows):
    ofport = ovs.get_ovs_patch_port_ofport(localnet)
    if ofport not in ovs_flows[bridge_device]['in_port']:
        ovs_flows[bridge_device]['in_port'].append(ofport)
        ovs.ensure_mac_tweak_flows(bridge_device,
                                   ovs_flows[bridge_device]['mac'],
                                   [ofport],
                                   constants.OVS_RULE_COOKIE)


def _wire_provider_port_underlay(routing_tables_routes, ovs_flows, port_ips,
                                 bridge_device, bridge_vlan, localnet,
                                 routing_table, proxy_cidrs, lladdr=None):
    if not bridge_device:
        return False
    for ip in port_ips:
        try:
            if lladdr:
                dev = bridge_device
                if bridge_vlan:
                    dev = '{}.{}'.format(dev, bridge_vlan)
                linux_net.add_ip_rule(ip, routing_table[bridge_device],
                                      dev=dev, lladdr=lladdr)
            else:
                linux_net.add_ip_rule(ip, routing_table[bridge_device],
                                      dev=bridge_device)
        except agent_exc.InvalidPortIP:
            LOG.exception("Invalid IP to create a rule for port on the "
                          "provider network: %s", ip)
            return False
        linux_net.add_ip_route(routing_tables_routes, ip,
                               routing_table[bridge_device], bridge_device,
                               vlan=bridge_vlan)
    # add proxy ndp config for ipv6
    for n_cidr in proxy_cidrs:
        if linux_net.get_ip_version(n_cidr) == constants.IP_VERSION_6:
            linux_net.add_ndp_proxy(n_cidr, bridge_device, bridge_vlan)
    # NOTE(ltomasbo): This is needed as the patch ports are not created
    # until the first VM/FIP in that provider network is created in a node
    try:
        _ensure_updated_mac_tweak_flows(localnet, bridge_device, ovs_flows)
    except agent_exc.PatchPortNotFound:
        LOG.warning("Patch port %s for bridge %s not found. Not possible to "
                    "create the needed ovs flows for the outgoing traffic. "
                    "It will be retried at the resync.", localnet,
                    bridge_device)
        return False
    return True


def _wire_provider_port_evpn(routing_tables_routes, ovs_flows, port_ips,
                             bridge_device, bridge_vlan, localnet,
                             proxy_cidrs, mac=None, via=None):
    try:
        evpn_dev = evpn.lookup_vlan(bridge_device, bridge_vlan)
    except KeyError:
        msg = ('EVPN has not been setup for bridge %s with vlan device %s. '
               'Either the network has not been configured, or something '
               'went wrong in the base wiring method.')
        LOG.warning(msg, bridge_device, bridge_vlan)
        return

    via = via or {}  # make sure it is at least a empty dictionary

    for ip in port_ips:
        ver = linux_net.get_ip_version(ip)
        evpn_dev.add_route(routing_tables_routes, ip, mac, via=via.get(ver))

    return True


def _wire_provider_port_ovn(ovn_idl, port_ips, mac):
    cmds = []
    port = "{}-openstack".format(constants.OVN_CLUSTER_ROUTER)
    for port_ip in port_ips:
        cmds.append(ovn_idl.static_mac_binding_add(
            port, port_ip, mac, override_dynamic_mac=True, may_exist=True))
    if cmds:
        _execute_commands(ovn_idl, cmds)
    # to keep it consisten with the underlay method
    return True


def _unwire_provider_port_underlay(routing_tables_routes, port_ips,
                                   bridge_device, bridge_vlan, routing_table,
                                   proxy_cidrs, lladdr=None):
    if not bridge_device:
        return False
    for ip in port_ips:
        if lladdr:
            if linux_net.get_ip_version(ip) == constants.IP_VERSION_6:
                cr_lrp_ip = '{}/128'.format(ip)
            else:
                cr_lrp_ip = '{}/32'.format(ip)
            try:
                dev = bridge_device
                if bridge_vlan:
                    dev = '{}.{}'.format(dev, bridge_vlan)
                linux_net.del_ip_rule(cr_lrp_ip, routing_table[bridge_device],
                                      dev=dev, lladdr=lladdr)
            except agent_exc.InvalidPortIP:
                LOG.exception("Invalid IP to delete a rule for the "
                              "provider port: %s", cr_lrp_ip)
                return False
        else:
            try:
                linux_net.del_ip_rule(ip, routing_table[bridge_device],
                                      dev=bridge_device)
            except agent_exc.InvalidPortIP:
                LOG.exception("Invalid IP to delete a rule for the "
                              "provider port: %s", ip)
                return False
        linux_net.del_ip_route(routing_tables_routes, ip,
                               routing_table[bridge_device], bridge_device,
                               vlan=bridge_vlan)
    for n_cidr in proxy_cidrs:
        if linux_net.get_ip_version(n_cidr) == constants.IP_VERSION_6:
            linux_net.del_ndp_proxy(n_cidr, bridge_device, bridge_vlan)
    return True


def _unwire_provider_port_evpn(routing_tables_routes, port_ips,
                               bridge_device, bridge_vlan, lladdr):
    # locate the evpn_dev, based on bridge and vlan
    try:
        evpn_dev = evpn.lookup_vlan(bridge_device, bridge_vlan)
    except KeyError:
        msg = ('EVPN has not been setup for bridge %s with vlan device %s. '
               'Either the network has not been configured, or something '
               'went wrong in the base wiring method.')
        LOG.warning(msg, bridge_device, bridge_vlan)
        return

    for ip in port_ips:
        evpn_dev.del_route(routing_tables_routes, ip, lladdr)

    return True


def _unwire_provider_port_ovn(ovn_idl, port_ips):
    cmds = []
    port = "{}-openstack".format(constants.OVN_CLUSTER_ROUTER)
    for port_ip in port_ips:
        cmds.append(ovn_idl.static_mac_binding_del(
            port, port_ip, if_exists=True))
    if cmds:
        _execute_commands(ovn_idl, cmds)
    # to keep it consisten with the underlay method
    return True


def wire_lrp_port(routing_tables_routes, ip, bridge_device, bridge_vlan,
                  routing_tables, cr_lrp_ips):
    if CONF.exposing_method == constants.EXPOSE_METHOD_UNDERLAY:
        return _wire_lrp_port_underlay(routing_tables_routes, ip,
                                       bridge_device, bridge_vlan,
                                       routing_tables, cr_lrp_ips)
    elif CONF.exposing_method == constants.EXPOSE_METHOD_VRF:
        return _wire_lrp_port_evpn(routing_tables_routes, ip, bridge_device,
                                   bridge_vlan, cr_lrp_ips)
    elif CONF.exposing_method == constants.EXPOSE_METHOD_OVN:
        # TODO(ltomasbo): Add flow on br-ex(-X)
        # ovs-ofctl add-flow br-ex
        # "cookie=0xbadcaf2,ip,nw_dst=20.0.0.0/24,in_port=enp2s0,priority=100,
        # actions=mod_dl_dst:$ENP2S0_MAC,output=$patch"
        # Add router route to go through cr-lrp ip:
        # ovn-nbctl lr-route-add bgp-router 20.0.0.0/24 172.16.100.143
        #     bgp-router-public
        return


def _wire_lrp_port_underlay(routing_tables_routes, ip, bridge_device,
                            bridge_vlan, routing_tables, cr_lrp_ips):
    if not bridge_device:
        return False
    LOG.debug("Adding IP Rules for network %s", ip)
    try:
        linux_net.add_ip_rule(ip, routing_tables[bridge_device])
    except agent_exc.InvalidPortIP:
        LOG.exception("Invalid IP to create a rule for the lrp (network "
                      "router interface) port: %s", ip)
        return False
    LOG.debug("Added IP Rules for network %s", ip)

    LOG.debug("Adding IP Routes for network %s", ip)
    # NOTE(ltomasbo): This assumes the provider network can only have
    # (at most) 2 subnets, one for IPv4, one for IPv6
    ip_version = linux_net.get_ip_version(ip)
    for cr_lrp_ip in cr_lrp_ips:
        if linux_net.get_ip_version(cr_lrp_ip) == ip_version:
            linux_net.add_ip_route(
                routing_tables_routes,
                ip.split("/")[0],
                routing_tables[bridge_device],
                bridge_device,
                vlan=bridge_vlan,
                mask=ip.split("/")[1],
                via=cr_lrp_ip)

            if (CONF.advertisement_method_tenant_networks ==
                    constants.ADVERTISEMENT_METHOD_SUBNET):
                # NOTE(jayjahns): The route for the tenant subnet needs to be
                # added to the bgp_nic in order to announce the network.
                linux_net.add_ip_route(
                    routing_tables_routes,
                    ip.split("/")[0],
                    CONF.bgp_vrf_table_id,
                    CONF.bgp_nic,
                    mask=ip.split("/")[1],
                    via=cr_lrp_ip)
            break
    LOG.debug("Added IP Routes for network %s", ip)
    return True


def _wire_lrp_port_evpn(routing_tables_routes, ip, bridge_device,
                        bridge_vlan, cr_lrp_ips):

    # Generate the via addresses
    via = driver_utils.ips_per_version(cr_lrp_ips)

    try:
        evpn_dev = evpn.lookup_vlan(bridge_device, bridge_vlan)
    except KeyError:
        msg = ('EVPN has not been setup for bridge %s with vlan device %s. '
               'Either the network has not been configured, or something '
               'went wrong in the base wiring method.')
        LOG.warning(msg, bridge_device, bridge_vlan)
        return

    ver = linux_net.get_ip_version(ip)
    evpn_dev.add_route(routing_tables_routes, ip, None, via=via.get(ver))
    return True


def unwire_lrp_port(routing_tables_routes, ip, bridge_device, bridge_vlan,
                    routing_tables, cr_lrp_ips):
    if CONF.exposing_method == constants.EXPOSE_METHOD_UNDERLAY:
        return _unwire_lrp_port_underlay(routing_tables_routes, ip,
                                         bridge_device, bridge_vlan,
                                         routing_tables, cr_lrp_ips)
    elif CONF.exposing_method == constants.EXPOSE_METHOD_VRF:
        return _unwire_lrp_port_evpn(routing_tables_routes, ip,
                                     bridge_device, bridge_vlan)
    elif CONF.exposing_method == constants.EXPOSE_METHOD_OVN:
        # TODO(ltomasbo): Remove flow(s) and router route
        return


def _unwire_lrp_port_underlay(routing_tables_routes, ip, bridge_device,
                              bridge_vlan, routing_tables, cr_lrp_ips):
    if not bridge_device:
        return False
    LOG.debug("Deleting IP Rules for network %s", ip)
    try:
        linux_net.del_ip_rule(ip, routing_tables[bridge_device])
    except agent_exc.InvalidPortIP:
        LOG.exception("Invalid IP to delete a rule for the "
                      "lrp (network router interface) port: %s", ip)
        return False
    LOG.debug("Deleted IP Rules for network %s", ip)

    LOG.debug("Deleting IP Routes for network %s", ip)
    ip_version = linux_net.get_ip_version(ip)
    for cr_lrp_ip in cr_lrp_ips:
        if linux_net.get_ip_version(cr_lrp_ip) == ip_version:
            linux_net.del_ip_route(
                routing_tables_routes,
                ip.split("/")[0],
                routing_tables[bridge_device],
                bridge_device,
                vlan=bridge_vlan,
                mask=ip.split("/")[1],
                via=cr_lrp_ip)

            if (CONF.advertisement_method_tenant_networks ==
                    constants.ADVERTISEMENT_METHOD_SUBNET):
                # NOTE(jayjahns): We need to delete the route to the
                # tenant subnet from the bgp_nic interface.
                linux_net.del_ip_route(
                    routing_tables_routes,
                    ip.split("/")[0],
                    CONF.bgp_vrf_table_id,
                    CONF.bgp_nic,
                    mask=ip.split("/")[1],
                    via=cr_lrp_ip)
    LOG.debug("Deleted IP Routes for network %s", ip)
    return True


def _unwire_lrp_port_evpn(routing_tables_routes, ip, bridge_device,
                          bridge_vlan):
    # locate the evpn_dev, based on bridge and vlan
    try:
        evpn_dev = evpn.lookup_vlan(bridge_device, bridge_vlan)
    except KeyError:
        msg = ('EVPN has not been setup for bridge %s with vlan device %s. '
               'Either the network has not been configured, or something '
               'went wrong in the base wiring method.')
        LOG.warning(msg, bridge_device, bridge_vlan)
        return

    evpn_dev.del_route(routing_tables_routes, ip)

    return True
